import itertools
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from typing import (
    Any,
    FrozenSet,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
)

import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore
from sklearn.linear_model import LogisticRegression  # type: ignore

from ..activelearning.base import ActiveLearner
from ..activelearning.ml_based import FeatureMatrix, MLBased
from ..environment import AbstractEnvironment
from ..machinelearning import AbstractClassifier
from ..typehints import DT, IT, KT, LT, RT, VT


@dataclass
class Estimate:
    point: float
    lower_bound: float
    upper_bound: float


class AbstractEstimator(ABC, Generic[IT, KT, DT, VT, RT, LT]):
    name = "AbstractEstimator"

    @abstractmethod
    def __call__(
        self, learner: ActiveLearner[IT, KT, DT, VT, RT, LT], label: LT
    ) -> Estimate:
        raise NotImplementedError


@dataclass
class DecisionRow(Generic[KT]):
    key: KT
    probability: float
    labeled: bool
    pos_labeled: bool
    order: Optional[int]



from typing import Any

import numpy as np
import numpy.typing as npt
from scipy.stats import entropy  # type:ignore

from .catalog import ALCatalog
from .selectioncriterion import AbstractSelectionCriterion


class MostCertainSampling(AbstractSelectionCriterion):
    """Selects the training examples most far away from the
    decision threshold at 0.50 / 50 % class probability.
    """

    name = ALCatalog.QueryType.MOST_CERTAIN

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        min_prob: npt.NDArray[Any] = np.amin(np.abs(prob_vec - 0.5), axis=1)  # type: ignore
        return min_prob


class MinEntropy(AbstractSelectionCriterion):
    """Selects the training examples with the lowest entropy
    at the probability level. This method is usable for
    Multilabel Classification.
    """

    name = ALCatalog.QueryType.MIN_ENTROPY

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        min_entropy: npt.NDArray[Any] = -entropy(prob_mat, axis=1)  # type: ignore
        return min_entropy


class MostConfidence(AbstractSelectionCriterion):
    """Selects the training examples with the highest probability
    for **any** label in the probability matrix.
    """

    name = ALCatalog.QueryType.MOST_CONFIDENCE

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        confidence: npt.NDArray[Any] = np.amax(prob_mat, axis=1)  # type: ignore
        return confidence


class LabelMaximizer(AbstractSelectionCriterion):
    """Identity function. This is usable for finding the
    instance with the highest probability when the matrix is
    sliced for one label only.

    TODO: Make the label a parameter
    """

    name = ALCatalog.QueryType.LABELMAXIMIZER

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        return prob_mat


class LabelMaximizerNew(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.LABELMAXIMIZER_NEW

    def __init__(self, label_column: int):
        super().__init__()
        self.label_column = label_column

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        prob_mat_sliced = prob_mat[:, self.label_column]
        return prob_mat_sliced



import itertools

from typing import Optional, Tuple, TypeVar, Sequence, Iterator, List, Iterable
from operator import itemgetter
_T = TypeVar("_T")

def divide_sequence(full_list: Sequence[_T], batch_size: int) -> Iterator[Sequence[_T]]: 
    # looping till length l 
    for i in range(0, len(full_list), batch_size):  
        yield full_list[i:i + batch_size] 

def divide_iterable(iterable: Iterable[_T], batch_size: int) -> Iterator[Iterator[_T]]:
    iterator = iter(iterable)
    for first in iterator:
        yield itertools.chain([first], itertools.islice(iterator, batch_size - 1))

def divide_iterable_in_lists(iterable: Iterable[_T], batch_size: int) -> Iterator[Sequence[_T]]:
    return map(list, divide_iterable(iterable, batch_size))


def get_consecutive(iterable: Sequence[int]) -> Iterable[Sequence[int]]:
    results = [list(map(itemgetter(1), g)) for k, g in itertools.groupby(enumerate(iterable), lambda x: x[0]-x[1])]
    yield from results # type: ignore

def get_range(iterable: Sequence[int]) -> Iterable[Tuple[int, Optional[int]]]:
    def minmax(sorted_iterable: Iterable[int]) -> Tuple[int, Optional[int]]:
        int_list = list(sorted_iterable)
        assert len(int_list) > 0
        min_int = int_list[0]
        max_int = int_list[-1]
        if min_int == max_int:
            return min_int, None
        return min_int, max_int + 1
    map_result = map(minmax, get_consecutive(iterable))
    yield from map_result


from typing import Any, Iterable, Iterator, Optional, Sequence, Tuple, TypeVar, Union

from h5py._hl.dataset import Dataset # type: ignore

import numpy as np
import numpy.typing as npt
from .func import list_unzip
import itertools

KT = TypeVar("KT")
VT = TypeVar("VT")

def get_lists(slices: Iterable[Tuple[int, Optional[int]]]) -> Sequence[int]:
    def convert_back(slice: Tuple[int, Optional[int]]) -> Sequence[int]:
        start, end = slice
        if end is None:
            return [start]
        idxs = list(range(start, end))
        return idxs

    result = list(itertools.chain.from_iterable(map(convert_back, slices)))
    return result


def memslicer(matrix: Union[Dataset, npt.NDArray[Any]], slices: Iterable[Tuple[int, Optional[int]]]) -> npt.NDArray[Any]:
    idxs = get_lists(slices)
    min_idx, max_idx= min(idxs), max(idxs)
    new_idxs = tuple([idx - min_idx for idx in idxs])
    dims = len(matrix.shape)
    
    def get_slice_1d():
        big_slice_mat = matrix[min_idx:(max_idx + 1)]        
        small_slice_mat = big_slice_mat[new_idxs] # type: ignore
        return small_slice_mat
    def get_slice_2d():
        big_slice_mat = matrix[min_idx:(max_idx + 1),:]
        small_slice_mat = big_slice_mat[new_idxs, :] # type: ignore
        return small_slice_mat
   
    if dims == 1:
        mat = get_slice_1d()
        return mat
    if dims == 2:
        mat = get_slice_2d()
        return mat
    raise NotImplementedError("No Slicing for 3d yet")


def slicer(matrix: Union[Dataset, npt.NDArray[Any]], slices: Iterable[Tuple[int, Optional[int]]]) -> npt.NDArray[Any]:
        def get_slices_1d():
            for slice_min, slice_max in slices:
                if slice_max is not None:
                    yield matrix[slice_min:slice_max]
                else:
                    yield matrix[slice_min]
        def get_slices_2d():
            for slice_min, slice_max in slices:
                if slice_max is not None:
                    yield matrix[slice_min:slice_max,:]
                else:
                    yield matrix[slice_min,:]
        dims = len(matrix.shape)
        if dims == 1:
            return np.hstack(list(get_slices_1d())) # type: ignore
        return np.vstack(list(get_slices_2d())) # type: ignore

def matrix_to_vector_list(matrix: npt.NDArray[Any]) -> Sequence[npt.NDArray[Any]]:
    def get_vector(index: int) -> npt.NDArray[Any]:
        return matrix[index, :]
    n_rows = matrix.shape[0]
    rows = range(n_rows)
    return list(map(get_vector, rows))

def matrix_tuple_to_vectors(keys: Sequence[KT], 
                            matrix: npt.NDArray[Any]
                           ) -> Tuple[Sequence[KT], Sequence[npt.NDArray[Any]]]:
    return keys, matrix_to_vector_list(matrix)

def matrix_tuple_to_zipped(keys: Sequence[KT], 
                           matrix: npt.NDArray[Any]) -> Sequence[Tuple[KT, npt.NDArray[Any]]]:
    result = list(zip(keys, matrix_to_vector_list(matrix)))
    return result

def raw_proba_chainer(itera: Iterator[Tuple[Sequence[KT], npt.NDArray[Any]]]) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
    key_lists, matrices = list_unzip(itera)
    keys = list(itertools.chain(*key_lists))
    matrix = np.vstack(matrices)
    return keys, matrix


import itertools
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

import numpy as np
import numpy.typing as npt
from instancelib.utils.func import flatten_dicts, list_unzip, value_map

from allib.analysis.initialization import (
    Initializer,
    RandomInitializer,
    SeparateInitializer,
)
from ..estimation.autostop import HorvitzThompsonVar2

from ..estimation.base import AbstractEstimator
from ..estimation.mhmodel import AbundanceEstimator
from ..estimation.rasch_comb_parametric import EMRaschRidgeParametricPython
from ..estimation.rasch_multiple import EMRaschRidgeParametricConvPython
from ..estimation.rasch_parametric import ParametricRaschPython
from ..estimation.rasch_python import EMRaschRidgePython
from ..stopcriterion.base import AbstractStopCriterion
from ..stopcriterion.catalog import StopCriterionCatalog
from ..stopcriterion.estimation import (
    CombinedStopCriterion,
    Conservative,
    Optimistic,
    UpperboundCombinedCritertion,
)
from ..stopcriterion.heuristic import AprioriRecallTarget
from ..stopcriterion.others import (
    BudgetStoppingRule,
    KneeStoppingRule,
    ReviewHalfStoppingRule,
    Rule2399StoppingRule,
    StopAfterKNegative,
)
from ..typehints import LT
from .catalog import (
    ALConfiguration,
    EstimationConfiguration,
    ExperimentCombination,
    FEConfiguration,
    StopBuilderConfiguration,
)
from .ensemble import (
    al_config_ensemble_prob,
    al_config_entropy,
    naive_bayes_estimator,
    rasch_estimator,
    rasch_lr,
    rasch_nblrrf,
    rasch_nblrrflgbm,
    rasch_nblrrflgbmrand,
    rasch_nblrrfsvm,
    rasch_rf,
    svm_estimator,
    tf_idf5000,
)

from .tarbaselines import autotar, autostop

_K = TypeVar("_K")
_T = TypeVar("_T")
_U = TypeVar("_U")

AL_REPOSITORY = {
    ALConfiguration.NaiveBayesEstimator: naive_bayes_estimator,
    ALConfiguration.SVMEstimator: svm_estimator,
    ALConfiguration.RaschEstimator: rasch_estimator,
    ALConfiguration.EntropySamplingNB: al_config_entropy,
    ALConfiguration.ProbabilityEnsemble: al_config_ensemble_prob,
    ALConfiguration.RaschLR: rasch_lr,
    ALConfiguration.RaschNBLRRF: rasch_nblrrf,
    ALConfiguration.RaschNBLRRFSVM: rasch_nblrrfsvm,
    ALConfiguration.RaschRF: rasch_rf,
    ALConfiguration.RaschNBLRRFLGBM: rasch_nblrrflgbm,
    ALConfiguration.RaschNBLRRFLGBMRAND: rasch_nblrrflgbmrand,
    ALConfiguration.AUTOTAR: autotar,
    ALConfiguration.AUTOSTOP: autostop,
}

FE_REPOSITORY = {FEConfiguration.TFIDF5000: tf_idf5000}

ESTIMATION_REPOSITORY = {
    EstimationConfiguration.RaschRidge: EMRaschRidgePython[
        int, str, npt.NDArray[Any], str, str
    ](),
    EstimationConfiguration.RaschParametric: ParametricRaschPython[
        int, str, npt.NDArray[Any], str, str
    ](),
    EstimationConfiguration.RaschApproxParametric: EMRaschRidgeParametricPython[
        int, str, npt.NDArray[Any], str, str
    ](),
    EstimationConfiguration.RaschApproxConvParametric: EMRaschRidgeParametricConvPython[
        int, str, npt.NDArray[Any], str, str
    ](),
    EstimationConfiguration.CHAO: AbundanceEstimator[Any, Any, Any, Any, Any, str](),
    EstimationConfiguration.AUTOSTOP: HorvitzThompsonVar2(),
}


def filter_mapping(mapping: Mapping[_K, Optional[_T]]) -> Mapping[_K, _T]:
    return {k: v for k, v in mapping.items() if v is not None}


def key_map(f: Callable[[_K], _U], mapping: Mapping[_K, _T]) -> Mapping[_U, _T]:
    return {f(k): v for k, v in mapping.items()}


def mapping_unzip(
    mapping: Mapping[_K, Tuple[_T, _U]]
) -> Tuple[Mapping[_K, _T], Mapping[_K, _U]]:
    left_dict = {k: v for k, (v, _) in mapping.items()}
    right_dict = {k: v for k, (_, v) in mapping.items()}
    return left_dict, right_dict


@dataclass(frozen=True)
class TarExperimentParameters(Generic[LT]):
    al_configuration: ALConfiguration
    fe_configuration: FEConfiguration
    init_configuration: Callable[..., Initializer[Any, Any, LT]]
    stop_builder_configuration: Sequence[StopBuilderConfiguration]
    batch_size: int
    stop_interval: int
    estimation_interval: int


def conservative_optimistic_builder(
    estimators: Mapping[str, AbstractEstimator], targets: Sequence[float]
) -> Callable[
    [LT, LT],
    Tuple[Mapping[str, AbstractEstimator], Mapping[str, AbstractStopCriterion[LT]]],
]:
    def builder(
        pos_label: LT, neg_label: LT
    ) -> Tuple[
        Mapping[str, AbstractEstimator], Mapping[str, AbstractStopCriterion[LT]]
    ]:
        conservatives = {
            f"{key}_conservative_{target}": Conservative.builder(est, target)(
                pos_label, neg_label
            )
            for key, est in estimators.items()
            for target in targets
        }
        optimistics = {
            f"{key}_optimistic_{target}": Optimistic.builder(est, target)(
                pos_label, neg_label
            )
            for key, est in estimators.items()
            for target in targets
        }
        return estimators, flatten_dicts(conservatives, optimistics)

    return builder


def standoff_builder(
    pos_label: LT, neg_label: LT
) -> Tuple[Mapping[str, AbstractEstimator], Mapping[str, AbstractStopCriterion[LT]]]:
    recall95 = AprioriRecallTarget(pos_label, 0.95)
    recall100 = AprioriRecallTarget(pos_label, 1.0)
    knee = KneeStoppingRule(pos_label)
    half = ReviewHalfStoppingRule(pos_label)
    budget = BudgetStoppingRule(pos_label)
    rule2399 = Rule2399StoppingRule(pos_label)
    stop200 = StopAfterKNegative(pos_label, 200)
    stop400 = StopAfterKNegative(pos_label, 400)
    criteria = {
        "Perfect95": recall95,
        "Perfect100": recall100,
        "Half": half,
        "Knee": knee,
        "Budget": budget,
        "Rule2399": rule2399,
        "Stop200": stop200,
        "Stop400": stop400,
    }
    return dict(), criteria


TARGETS = [0.7, 0.8, 0.9, 0.95, 1.0]

STOP_BUILDER_REPOSITORY = {
    StopBuilderConfiguration.CHAO_CONS_OPT: conservative_optimistic_builder(
        {"Chao": AbundanceEstimator()}, TARGETS
    ),
    StopBuilderConfiguration.AUTOTAR: standoff_builder,
    StopBuilderConfiguration.AUTOSTOP: conservative_optimistic_builder(
        {"AUTOSTOP": HorvitzThompsonVar2()}, TARGETS
    ),
}


EXPERIMENT_REPOSITORY = {
    ExperimentCombination.CHAO4: TarExperimentParameters(
        ALConfiguration.RaschNBLRRFLGBMRAND,
        FEConfiguration.TFIDF5000,
        SeparateInitializer.builder(1),
        (StopBuilderConfiguration.CHAO_CONS_OPT,),
        10,
        10,
        10,
    ),
    ExperimentCombination.AUTOTAR: TarExperimentParameters(
        ALConfiguration.AUTOTAR,
        FEConfiguration.TFIDF5000,
        RandomInitializer.builder(5),
        (StopBuilderConfiguration.AUTOTAR,),
        10,
        10,
        10,
    ),
    ExperimentCombination.AUTOSTOP: TarExperimentParameters(
        ALConfiguration.AUTOSTOP,
        FEConfiguration.TFIDF5000,
        RandomInitializer.builder(5),
        (StopBuilderConfiguration.AUTOSTOP,),
        10,
        10,
        10,
    ),
}



import collections
from typing import Any, Deque, FrozenSet, Generic, Mapping, Sequence
from ..analysis.base import AbstractStatistics, StatsMixin, AnnotationStatistics
from ..activelearning.base import ActiveLearner
from ..typehints.typevars import KT, LT
from .base import AbstractStopCriterion
from scipy.stats import hypergeom
import numpy as np



class StatsStoppingCriterion(AbstractStopCriterion[LT], Generic[KT, LT]):
    stats: AbstractStatistics[KT, LT]
    def __init__(self, pos_label: LT) -> None:
        super().__init__()
        self.pos_label = pos_label
        self.stats = AnnotationStatistics[KT, LT]()
    
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        if isinstance(learner, StatsMixin):
            self.stats = learner.stats
        else:
            self.stats.update(learner)
        

class StopAfterKNegative(StatsStoppingCriterion[KT, LT]):
    def __init__(self, pos_label: LT, k: int) -> None:
        super().__init__(pos_label)
        self.k = k
    
    @property
    def stop_criterion(self) -> bool:
        annotated_since_last_pos = self.stats.annotations_since_last(self.pos_label)
        return annotated_since_last_pos >= self.k
        

class KneeStoppingRule(StatsStoppingCriterion[KT, LT]):
    """
    .. seealso::
        .. [1] Gordon V. Cormack, and Maura R. Grossman. "Engineering quality and reliability in technology-assisted review." 
               *Proceedings of the 39th International ACM SIGIR conference on Research and Development in Information Retrieval.* 2016.
               `<https://dl.acm.org/doi/10.1145/2911451.2911510>`__
    """
    @property
    def stop_criterion(self) -> bool:
        if self.stats.rounds < 1:
            return False

        pos_per_round = np.array(self.stats.label_per_round(self.pos_label))
        pos_found = pos_per_round.cumsum()
        
        rho_s = -1
        for i in range(self.stats.rounds):
            rho = (pos_found[i]/(i+1)) / ((1+pos_found[-1]-pos_found[i])/(self.stats.rounds-i))
            rho_s = max(rho_s, rho)

        return rho_s >= 156 - min(pos_found[-1], 150)


class BudgetStoppingRule(StatsStoppingCriterion[KT, LT]):
    """
    .. seealso::
        .. [2] Gordon V. Cormack, and Maura R. Grossman. "Engineering quality and reliability in technology-assisted review." 
               *Proceedings of the 39th International ACM SIGIR conference on Research and Development in Information Retrieval.* 2016.
               `<https://dl.acm.org/doi/10.1145/2911451.2911510>`__
    """
    @property
    def stop_criterion(self) -> bool:
        if self.stats.rounds < 1:
            return False
            
        batchsize = self.stats.annotations_per_round[-1]
        pos_per_round = np.array([self.stats.label_per_round(self.pos_label)])
        pos_found = pos_per_round.cumsum()
        n_docs = self.stats.dataset_size
        n_rounds = self.stats.rounds
        rho_s = -1
        for i in range(n_rounds):
            rho = (pos_found[i]/(i+1)) / ((1+pos_found[-1]-pos_found[i])/(n_rounds-i))
            rho_s = max(rho_s, rho)

        return  (rho_s >= 6 and batchsize*self.stats.rounds+1 >= 10 * n_docs / pos_found[n_rounds - 1]) or \
                (self.stats.current_annotated >= n_docs*0.75)


class ReviewHalfStoppingRule(StatsStoppingCriterion[KT, LT]):
    
    @property
    def stop_criterion(self) -> bool:
        if self.stats.rounds < 1:
            return False
        return self.stats.current_annotated >= self.stats.dataset_size // 2


class BatchPrecStoppingRule(StatsStoppingCriterion[KT, LT]):

    def __init__(self, pos_label: LT, prec_cutoff=5/200, slack=1) -> None:
        super().__init__(pos_label)
        self.prec_cutoff = prec_cutoff
        self.slack = slack
    
    @property
    def stop_criterion(self) -> bool:
        bprec = np.array([len(batch[self.pos_label]) / sum([len(batch[k]) for k in batch if k != self.pos_label]) for batch in self.stats.per_round])
        counter = 0
        for prec in bprec:
            counter = (counter+1) if prec <= self.prec_cutoff else 0
            if counter >= self.slack:
                return True
        return False


class Rule2399StoppingRule(StatsStoppingCriterion[KT, LT]):
    @property
    def stop_criterion(self) -> bool:
        return self.stats.current_annotated >= 1.2 * self.stats.current_label_count(self.pos_label) + 2399


# class QuantStoppingRule(StoppingRule):
#     """
#     .. seealso::
#         .. [3] Eugene Yang, David D. Lewis, and Ophir Frieder. "Heuristic stopping rules for technology-assisted review." 
#                *Proceedings of the 21st ACM Symposium on Document Engineering.* 2021.
#                `<https://arxiv.org/abs/2106.09871>`__
#     """
#     def __init__(self, target_recall: float, nstd: float = 0):
#         super().__init__(target_recall=target_recall)
#         self.nstd = nstd
    
#     def checkStopping(self, ledger: Ledger, workflow, **kwargs) -> bool:
#         if ledger.n_rounds < 2:
#             return False

#         scores = getOneDimScores(workflow.latest_scores)
            
#         assert (scores <= 1).all() and (scores >= 0).all(), \
#                 "Scores have to be probabilities to use Quant Rule."

#         # `ps` stands for probability sum
#         unknown_ps = scores[ ~ledger.annotated ].sum()
#         known_ps = scores[ ledger.annotated ].sum()
#         est_recall = (known_ps) / (known_ps + unknown_ps)
#         if self.nstd == 0:
#             return est_recall >= self.target_recall
        
#         prod = scores * (1-scores)
#         all_var = prod.sum()
#         unknown_var = prod[ ~ledger.annotated ].sum()

#         est_var = (known_ps**2 / (known_ps + unknown_ps)**4 * all_var) + (1 / (known_ps + unknown_ps)**2 * (all_var-unknown_var))
        
#         return est_recall - self.nstd*np.sqrt(est_var) >= self.target_recall

class CHMHeuristicsStoppingRule(StatsStoppingCriterion[KT, LT]):
    """
    .. seealso::
        .. [4] Max W. Callaghan, and Finn Müller-Hansen. "Statistical stopping criteria for automated screening in systematic reviews." 
               *Systematic Reviews 9.1* (2020): 1-14.
               `<https://pubmed.ncbi.nlm.nih.gov/33248464/>`__
    """
    def __init__(self, pos_label: LT, target_recall: float, alpha: float) -> None:
        super().__init__(pos_label)
        self.target_recall=target_recall
        self.alpha = alpha
    
    @property
    def stop_criterion(self) -> bool:
        if self.stats.rounds < 2:
            return False

        pos_per_round = np.array([self.stats.label_per_round(self.pos_label)])
        pos_found = pos_per_round.cumsum()
        annotated_cumsum = np.array([self.stats.annotations_per_round]).cumsum()
        n_docs = self.stats.dataset_size
        
        for i in range(1, self.stats.rounds):
            if hypergeom.cdf( pos_found[-1] - pos_found[i], # k
                              n_docs-annotated_cumsum[i], # N
                              int(pos_found[-1]/self.target_recall - pos_found[i]), # K_tar
                              annotated_cumsum[-1] - annotated_cumsum[i] # n
                            ) < self.alpha:
                return True
        return False



from __future__ import annotations
from argparse import Action

import collections
import functools
import itertools
import logging
from abc import ABC, abstractmethod
from multiprocessing import Pool
from queue import Queue
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

import numpy as np
import numpy.typing as npt
from instancelib.instances.base import Instance, InstanceProvider
from sklearn.exceptions import NotFittedError  # type: ignore

from ..environment import AbstractEnvironment
from ..exceptions import NoOrderingException, NotInitializedException
from ..exceptions.base import NoLabeledDataException, NoVectorsException
from ..machinelearning import AbstractClassifier
from ..utils import divide_sequence, mapsnd
from ..utils.func import filter_snd_none, list_unzip, sort_on
from .base import ActiveLearner
from .poolbased import PoolBasedAL
from .random import RandomSampling
from .selectioncriterion import AbstractSelectionCriterion

from ..typehints import KT, DT, VT, LT, RT, IT, LVT, PVT

FT = TypeVar("FT")

LOGGER = logging.getLogger(__name__)


class FeatureMatrix(Generic[KT]):
    def __init__(
        self, keys: Sequence[KT], vectors: Sequence[Optional[npt.NDArray[Any]]]
    ):
        # Filter all rows with None as Vector
        filtered_keys, filtered_vecs = filter_snd_none(keys, vectors)  # type: ignore
        self.matrix = np.vstack(filtered_vecs)
        self.indices: Sequence[KT] = filtered_keys

    def get_instance_id(self, row_idx: int) -> KT:
        return self.indices[row_idx]

    @classmethod
    def generator_from_provider_mp(
        cls,
        provider: InstanceProvider[Any, KT, Any, npt.NDArray[Any], Any],
        batch_size: int = 100,
    ) -> Iterator[FeatureMatrix[KT]]:
        for key_batch in divide_sequence(provider.key_list, batch_size):
            ret_keys, vectors = provider.bulk_get_vectors(key_batch)
            matrix = cls(ret_keys, vectors)
            yield matrix

    @classmethod
    def generator_from_provider(
        cls,
        provider: InstanceProvider[Any, KT, Any, npt.NDArray[Any], Any],
        batch_size: int = 100,
    ) -> Iterator[FeatureMatrix[KT]]:
        for tuple_batch in provider.vector_chunker(batch_size):
            keys, vectors = list_unzip(tuple_batch)
            matrix = cls(keys, vectors)
            yield matrix


class MLBased(
    PoolBasedAL[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT, LVT, PVT]
):
    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        classifier: AbstractClassifier[KT, VT, LT, LVT, PVT],
        fallback: ActiveLearner[IT, KT, DT, VT, RT, LT],
        batch_size=200,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        """Initialize an Machine Learning Based Active Learning method

        Parameters
        ----------
        classifier : AbstractClassifier[KT, VT, LT, LVT, PVT]
            The classifier
        fallback : PoolBasedAL[KT, DT, VT, RT, LT], optional
            If fitting a classifier is not possible choose
            this ActiveLearner for sampling,
            by default ``RandomSampling[KT, DT, VT, RT, LT]()``
        batch_size : int, optional
            The batch size for the feature matrix, by default 200
        """
        super().__init__(env, identifier=identifier)
        self.fitted = False
        self.classifier = classifier
        self.fallback = fallback
        self.batch_size = batch_size
        self._uses_fallback = False

    @property
    def uses_fallback(self) -> bool:
        """Returns true if the fallback method is used
        instead of the ML-based method

        Returns
        -------
        bool
            True if the model uses the fallback method
        """
        return self._uses_fallback

    def retrain(self) -> None:
        """Retrain the classifier using the information in the
        labeled set

        Raises
        ------
        NotInitializedException
            If the AL method has no attached Environment
        """
        LOGGER.info("[%s] Start with the retraining procedure", self.name)
        # Some sanity checks
        if self.env.labeled.empty:
            raise NoLabeledDataException(
                "There are no labeled instances for retraining"
            )

        # Collect the feature matrix for the labeled subset
        key_vector_pairs = itertools.chain.from_iterable(
            self.env.labeled.vector_chunker(self.batch_size)
        )
        keys, vectors = list_unzip(key_vector_pairs)
        if not vectors:
            raise NoVectorsException(
                "There are no vectors available for training the classifier"
            )
        LOGGER.info(
            "[%s] Gathered the feature matrix for all labeled documents", self.name
        )

        # Get all labels for documents in the labeled set
        labelings = list(map(self.env.labels.get_labels, keys))
        LOGGER.info("[%s] Gathered all labels", self.name)

        LOGGER.info("[%s] Start fitting the classifier", self.name)
        self.classifier.fit_vectors(vectors, labelings)
        LOGGER.info("[%s] Fitted the classifier", self.name)
        self.fitted = True

    def predict(
        self, instances: Sequence[Instance[KT, DT, VT, RT]]
    ) -> Sequence[FrozenSet[LT]]:
        """Predict the labels for the provided instances

        Parameters
        ----------
        instances : Sequence[Instance[KT, DT, VT, RT]]
            A sequence of instances. The instances should all
            have a vector of type `VT`; they should not be None

        Returns
        -------
        Sequence[FrozenSet[LT]]
            A list of labelings, matching the order of the `instances` parameters
        """
        return self.classifier.predict_instances(instances)

    def predict_proba(
        self, instances: Sequence[Instance[KT, DT, VT, RT]]
    ) -> Sequence[FrozenSet[Tuple[LT, float]]]:
        """Predict the labels and their probability
        for the provided instances

        Parameters
        ----------
        instances : Sequence[Instance[KT, DT, VT, RT]]
            A sequence of instances. The instances should all
            have a vector of type `VT`; they should not be None

        Returns
        -------
        Sequence[FrozenSet[Tuple[LT, float]]]
            A list of labelings and probabilities,
            matching the order of the `instances` parameters
        """
        return self.classifier.predict_proba_instances(instances)

    @property
    def name(self) -> Tuple[str, Optional[LT]]:
        return f"{self._name} :: {self.classifier.name}", None

    @staticmethod
    def iterator_fallback(func: Callable[..., IT]) -> Callable[..., IT]:
        """A decorator for fallback. If the ``__next__`` function
        fails, the fallback model's ``__next__`` will be used instead.

        The model catches the following exceptions:

        - :class:`sklearn.exceptions.NotFittedError` the model is not fitted
        - :class:`IndexError` e.g., there is no train data
        - :class:`ValueError` e.g., the train data has an incorrect format
        - :class:`StopIteration` e.g., there are no unlabeled documents
        - :class:`~allib.exceptions.base.NoOrderingException` an ordering
            could not be established

        Parameters
        ----------
        func : Callable[..., Instance[KT, DT, VT, RT]]
            The :meth:`__next__` method of a learner, that may fail

        Returns
        -------
        Callable[..., Instance[KT, DT, VT, RT]]
            The same method with an fallback
        """

        @functools.wraps(func)
        def wrapper(
            self: MLBased[IT, KT, DT, VT, RT, LT, LVT, PVT],
            *args: Any,
            **kwargs: Dict[str, Any],
        ) -> IT:
            if not self.uses_fallback:
                try:
                    return func(self, *args, **kwargs)
                except (
                    NotFittedError,
                    IndexError,
                    ValueError,
                    StopIteration,
                    NoOrderingException,
                    NoVectorsException,
                ) as ex:
                    LOGGER.error(
                        "[%s] Falling back to model %s, because of: %s",
                        self.name,
                        self.fallback.name,
                        ex,
                        exc_info=ex,
                    )
            LOGGER.warn(
                "[%s] Falling back to model %s, because it is not fitted",
                self.name,
                self.fallback.name,
            )
            fallback_value = next(self.fallback)
            return fallback_value

        return wrapper


class ProbabilityBased(
    MLBased[IT, KT, DT, npt.NDArray[Any], RT, LT, npt.NDArray[Any], npt.NDArray[Any]],
    ABC,
    Generic[IT, KT, DT, RT, LT],
):
    """
    An Active Learner that uses information in the probability matrix (i.e., the
    results of the classifier on the *unlabeled*  set of instances).

    Examples
    --------
    Usage:

    >>> classifier = SklearnClassifier(MultinomialNB(), LabelBinarizer())
    >>> sel_crit = EntropySampling() # Uncertainty Sampling
    >>> al = ProbabilityBased(classifier, sel_crit)
    >>> # Assume env contains your environment
    >>> al = al(env)
    >>> # Update the ordering
    >>> al.update_ordering()
    >>> # Retrieve the next document
    >>> instance = next(al)
    >>> # Label the document as Positive
    >>> label = "Positive"
    >>> al.env.labels.set_labels(instance, doc_label)
    >>> al.set_as_labeled(instance)
    """

    _name = "ProbabilityBased"

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT],
        classifier: AbstractClassifier[
            KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        selection_criterion: AbstractSelectionCriterion,
        fallback: ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT],
        batch_size=200,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(
            env, classifier, fallback, batch_size, *_, identifier=identifier, **__
        )
        self._selection_criterion = selection_criterion

    def selection_criterion(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        """Calculate the internal selection criterion for the given
        probability matrix

        Parameters
        ----------
        prob_mat : npt.NDArray[Any]
            The probability matrix with rows of class probabilities.
            Shape should be ``(n_instances, n_classes)``

        Returns
        -------
        npt.NDArray[Any]
            The result of the selection metrix. This has as shape
            ``(n_instances, )``
        """
        return self._selection_criterion(prob_mat)

    def _get_predictions(
        self, matrix: FeatureMatrix[KT]
    ) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
        """Calculate the probability matrix for the current feature matrix

        Args:
            matrix (FeatureMatrix[KT]): The matrix for which we want to know the predictions

        Returns:
            Tuple[Sequence[KT], npt.NDArray[Any]]: A list of keys and the probability predictions belonging to it
        """
        prob_vec: npt.NDArray[Any] = self.classifier.predict_proba(
            matrix.matrix
        )  # type: ignore
        keys = matrix.indices
        return keys, prob_vec

    # @ActiveLearner.ordering_log
    def calculate_ordering(self) -> Tuple[Sequence[KT], Sequence[float]]:
        """Calculate the ordering for the unlabeled set of instances according
        to their predicted label probabilities by the learner's classifier.
        The provided selection criterion calculates a metric on the probabilities.

        Returns
        -------
        Tuple[Sequence[KT], Sequence[float]]
            A tuple of two lists of equal length and ordered descendingly
            according to the second list.

            - A list of instance keys with type KT
            - A list of scores, matching with the instance keys
        """

        def get_metric_tuples(
            keys: Sequence[KT], vec: npt.NDArray[Any]
        ) -> Sequence[Tuple[KT, float]]:
            floats: Sequence[float] = vec.tolist()
            return list(zip(keys, floats))

        LOGGER.info(
            "[%s] Start calculating the prediction matrix for all unlabeled documents",
            self.name,
        )
        # Get a generator with that generates feature matrices from data
        matrices = FeatureMatrix[KT].generator_from_provider(
            self.env.unlabeled, self.batch_size
        )
        # Get the predictions for each matrix
        predictions = map(self._get_predictions, matrices)
        # Transfrorm the selection criterion function into a function that works on tuples and
        # applies the id :: a -> a function on the first element of the tuple and selection_criterion
        # on the second
        sel_func = mapsnd(self.selection_criterion)
        # Apply sel_func on the predictions
        metric_results = itertools.starmap(sel_func, predictions)
        # Transform the metric npt.NDArray[Any] to a python List[float] and flatten the iterable
        # to a list of Tuple[KT, float] where float is the metric for the instance with
        # key KT
        metric_tuples = list(
            itertools.chain.from_iterable(
                itertools.starmap(get_metric_tuples, metric_results)
            )
        )
        LOGGER.info("[%s] Calculated all metrics", self.name)
        # Sort the tuples in descending order, so that the key with the highest score
        # is on the first position of the list
        sorted_tuples = sort_on(1, metric_tuples)
        LOGGER.info("[%s] Sorted all metrics", self.name)
        # Retrieve the keys from the tuples
        ordered_keys, ordered_metrics = list_unzip(sorted_tuples)
        return ordered_keys, ordered_metrics

    def update_ordering(self) -> bool:
        """Calculates the ordering for Machine Learning based AL methods

        Returns
        -------
        bool
            True if updating succeeded
        """
        try:
            self.retrain()
            ordering, _ = self.calculate_ordering()
        except (
            NotFittedError,
            IndexError,
            ValueError,
            NoLabeledDataException,
            NoVectorsException,
        ) as ex:
            self._uses_fallback = True
            LOGGER.error(
                "[%s] Falling back to model %s, because of: %s",
                self.name,
                self.fallback.name,
                ex,
                exc_info=ex,
            )
            self.fallback.update_ordering()
            self._set_ordering([])
            return False
        else:
            self._uses_fallback = False
            self._set_ordering(ordering)
            return True

    @MLBased.iterator_fallback  # type: ignore
    def __next__(self) -> IT:
        value = super().__next__()
        return value

    @classmethod
    def builder(
        cls,
        classifier: AbstractClassifier[
            KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        selection_criterion: AbstractSelectionCriterion,
        fallback_builder: Callable[
            ..., ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]
        ],
        batch_size=200,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> Callable[..., ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT],]:
        def wrap_func(
            env: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT], 
            *_, **__
        ):
            built_classifier = classifier(env)
            fallback = fallback_builder(env)
            return cls(
                env,
                built_classifier,
                selection_criterion,
                fallback,
                batch_size=batch_size,
                identifier=identifier,
            )

        return wrap_func



from __future__ import annotations

from typing import Sequence, Optional, Any

import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator # type: ignore
from sklearn.exceptions import NotFittedError # type: ignore

from ..utils import SaveableInnerModel
from .base import BaseVectorizer


class SklearnVectorizer(BaseVectorizer[str], SaveableInnerModel[BaseEstimator]):
    innermodel: BaseEstimator
    _name = "SklearnVectorizer"

    def __init__(
        self,
        vectorizer: BaseEstimator,
        storage_location: Optional[str] = None,
        filename: Optional[str] = None
        ) -> None:
        BaseVectorizer.__init__(self) #type: ignore
        SaveableInnerModel.__init__(self, vectorizer, storage_location, filename)

    @SaveableInnerModel.load_model_fallback
    def fit(self, x_data: Sequence[str], **kwargs: Any) -> SklearnVectorizer:
        self.innermodel = self.innermodel.fit(x_data) # type: ignore
        self._fitted = True
        return self

    @SaveableInnerModel.load_model_fallback
    def transform(self, x_data: Sequence[str], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        if self.fitted:
            # TODO Check for performance issues with .toarray()
            return self.innermodel.transform(x_data).toarray() # type: ignore
        raise NotFittedError

    def fit_transform(self, x_data: Sequence[str], **kwargs: Any) -> npt.NDArray[Any]: # type: ignore
        self.fit(x_data)
        return self.transform(x_data) # type: ignore
    


import functools
from abc import ABC
from distutils.command.build import build
from typing import Any, Callable, Dict, List, Optional, Sequence, TypeVar

import instancelib as il
from instancelib.machinelearning.sklearn import SkLearnClassifier
from instancelib.typehints.typevars import DT, KT, LMT, LT, PMT, RT, VT
from sklearn.base import ClassifierMixin, TransformerMixin
from sklearn.linear_model import LogisticRegression

from ..environment.base import AbstractEnvironment
from ..factory import AbstractBuilder, ObjectFactory
from ..machinelearning import AbstractClassifier, MachineLearningFactory
from ..module.component import Component
from ..typehints.typevars import IT
from .autostop import AutoStopLearner
from .autotar import AutoTarLearner
from .base import ActiveLearner
from .catalog import ALCatalog as AL
from .ensembles import StrategyEnsemble
from .estimator import CycleEstimator, Estimator, RetryEstimator
from .labelmethods import LabelProbabilityBased
from .ml_based import ProbabilityBased
from .mostcertain import (LabelMaximizer, LabelMaximizerNew,
                          MostCertainSampling, MostConfidence)
from .prob_ensembles import (LabelMinProbEnsemble, LabelProbEnsemble,
                             ProbabilityBasedEnsemble)
from .random import RandomSampling
from .selectioncriterion import AbstractSelectionCriterion
from .uncertainty import (EntropySampling, LabelUncertainty,
                          LabelUncertaintyNew, LeastConfidence, MarginSampling,
                          NearDecisionBoundary, RandomMLStrategy)


class FallbackBuilder(AbstractBuilder):
    def __call__(self, **kwargs) -> Callable[[AbstractEnvironment], ActiveLearner]:
        if kwargs:
            fallback = self._factory.create(Component.ACTIVELEARNER, **kwargs)
            return fallback
        return RandomSampling.builder()


class ALBuilder(AbstractBuilder):
    def __call__(self, paradigm, **kwargs):
        return self._factory.create(paradigm, **kwargs)


class ProbabilityBasedBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        query_type: AL.QueryType,
        machinelearning: Dict,
        fallback: Dict = dict(),
        batch_size: int = 200,
        identifier: Optional[str] = None,
        **kwargs,
    ):
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        selection_criterion: AbstractSelectionCriterion = self._factory.create(
            query_type, **kwargs
        )
        built_fallback = self._factory.create(Component.FALLBACK, **fallback)
        return ProbabilityBased.builder(
            classifier,
            selection_criterion,
            built_fallback,
            batch_size=batch_size,
            identifier=identifier,
        )


class LabelProbabilityBasedBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        query_type: AL.QueryType,
        machinelearning: Dict,
        fallback: Dict = dict(),
        identifier: Optional[str] = None,
        **kwargs,
    ):
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        selection_criterion: AbstractSelectionCriterion = self._factory.create(
            query_type, **kwargs
        )
        built_fallback = self._factory.create(Component.FALLBACK, **fallback)
        return LabelProbabilityBased.builder(
            classifier,
            selection_criterion,
            built_fallback,            
            identifier=identifier,
        )


class PoolbasedBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self, query_type: AL.QueryType, identifier: Optional[str] = None, **kwargs
    ):
        return self._factory.create(query_type, identifier=identifier, **kwargs)

class CustomBuilder(AbstractBuilder):
    def __call__(self, method: AL.CustomMethods, **kwargs):
        return self._factory.create(method, **kwargs)


class StrategyEnsembleBuilder(AbstractBuilder):
    def build_learner(self, classifier: AbstractClassifier, config):
        query_type = config["query_type"]
        params = {k: v for k, v in config if k not in ["query_type"]}
        return self._factory.create(query_type, classifier=classifier, **params)

    def __call__(  # type: ignore
        self,
        learners: List[Dict],
        machinelearning: Dict,
        probabilities: List[float],
        identifier: Optional[str] = None,
        **kwargs,
    ):
        assert len(learners) == len(probabilities)
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        config_function = functools.partial(self.build_learner, classifier)
        configured_learners = list(map(config_function, learners))
        return StrategyEnsemble.builder(classifier, configured_learners, probabilities)


class CycleEstimatorBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self, learners: List[Dict], identifier: Optional[str] = None, **kwargs
    ) -> Callable[..., ActiveLearner]:
        configured_learners = [
            self._factory.create(Component.ACTIVELEARNER, **learner_config)
            for learner_config in learners
        ]
        return CycleEstimator.builder(configured_learners)


class EstimatorBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        learners: List[Dict],
        identifier: Optional[str] = None,
        **kwargs,
    ) -> Callable[..., ActiveLearner]:
        configured_learners = [
            self._factory.create(Component.ACTIVELEARNER, **learner_config)
            for learner_config in learners
        ]
        return Estimator.builder(configured_learners)


class RetryEstimatorBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self, learners: List[Dict], identifier: Optional[str] = None, **kwargs
    ) -> Callable[..., ActiveLearner]:
        configured_learners = [
            self._factory.create(Component.ACTIVELEARNER, **learner_config)
            for learner_config in learners
        ]
        return RetryEstimator.builder(configured_learners)


class SelectionCriterionBuilder(AbstractBuilder):
    def __call__(self, query_type: AL.QueryType, **kwargs):
        return self._factory.create(query_type, **kwargs)


class ProbabilityEnsembleBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        strategies: List[Dict],
        machinelearning: Dict,
        fallback: Dict = dict(),
        identifier: Optional[str] = None,
        **kwargs,
    ):
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        built_strategies: Sequence[AbstractSelectionCriterion] = [
            self._factory.create(Component.SELECTION_CRITERION, **dic)
            for dic in strategies
        ]
        built_fallback = self._factory.create(Component.FALLBACK, **fallback)
        return ProbabilityBasedEnsemble(
            classifier, built_strategies, fallback=built_fallback, identifier=identifier
        )


class LabelProbilityBasedEnsembleBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        strategy: AL.QueryType,
        machinelearning: Dict,
        fallback: Dict = dict(),
        identifier: Optional[str] = None,
        **kwargs,
    ):
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        if strategy not in self._factory.builders:
            raise NotImplementedError(
                f"The selection strategy {strategy} is not available"
            )
        chosen_strategy = self._factory.get_constructor(strategy)
        built_fallback = self._factory.create(Component.FALLBACK, **fallback)
        return LabelProbEnsemble(
            classifier, chosen_strategy, fallback=built_fallback, identifier=identifier
        )


class LabelMinProbilityBasedEnsembleBuilder(AbstractBuilder):
    def __call__(  # type: ignore
        self,
        strategy: AL.QueryType,
        machinelearning: Dict,
        fallback: Dict = dict(),
        identifier: Optional[str] = None,
        **kwargs,
    ):
        classifier = self._factory.create(Component.CLASSIFIER, **machinelearning)
        if strategy not in self._factory.builders:
            raise NotImplementedError(
                f"The selection strategy {strategy} is not available"
            )
        chosen_strategy = self._factory.get_constructor(strategy)
        built_fallback = self._factory.create(Component.FALLBACK, **fallback)
        return LabelMinProbEnsemble(
            classifier, chosen_strategy, fallback=built_fallback, identifier=identifier
        )


MT = TypeVar("MT")


def classifier_builder(
    classifier: MT,
    build_method: Callable[
        [MT, il.AbstractEnvironment[IT, KT, DT, VT, RT, LT]],
        il.AbstractClassifier[IT, KT, DT, VT, RT, LT, LMT, PMT],
    ],
) -> Callable[
    [il.AbstractEnvironment[IT, KT, DT, VT, RT, LT]],
    il.AbstractClassifier[IT, KT, DT, VT, RT, LT, LMT, PMT],
]:
    def wrap_func(env: il.AbstractEnvironment[IT, KT, DT, VT, RT, LT]):
        return build_method(classifier, env)

    return wrap_func


class AutoTARBuilder(AbstractBuilder):
    def __call__(self, k_sample: int, batch_size: int, **kwargs):
        logreg = LogisticRegression(solver="lbfgs", C=1.0, max_iter=10000)
        builder = classifier_builder(logreg, il.SkLearnVectorClassifier.build)
        at = AutoTarLearner.builder(builder, k_sample, batch_size, **kwargs)
        return at

class AutoSTOPBuilder(AbstractBuilder):
    def __call__(self, k_sample: int, batch_size: int, **kwargs):
        logreg = LogisticRegression(solver="lbfgs", C=1.0, max_iter=10000)
        builder = classifier_builder(logreg, il.SkLearnVectorClassifier.build)
        at = AutoStopLearner.builder(builder, k_sample, batch_size, **kwargs)
        return at


class ActiveLearningFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.attach(MachineLearningFactory())

        self.register_builder(Component.ACTIVELEARNER, ALBuilder())
        self.register_builder(Component.FALLBACK, FallbackBuilder())
        self.register_builder(
            Component.SELECTION_CRITERION, SelectionCriterionBuilder()
        )
        self.register_builder(AL.Paradigm.POOLBASED, PoolbasedBuilder())
        self.register_builder(AL.Paradigm.PROBABILITY_BASED, ProbabilityBasedBuilder())
        self.register_builder(AL.Paradigm.ESTIMATOR, EstimatorBuilder())
        self.register_builder(AL.Paradigm.CYCLE_ESTIMATOR, CycleEstimatorBuilder())
        self.register_builder(AL.Paradigm.CUSTOM, CustomBuilder())
        self.register_builder(AL.Paradigm.ENSEMBLE, StrategyEnsembleBuilder())
        self.register_builder(
            AL.Paradigm.LABEL_PROBABILITY_BASED, LabelProbabilityBasedBuilder()
        )
        self.register_builder(
            AL.Paradigm.PROBABILITY_BASED_ENSEMBLE, ProbabilityEnsembleBuilder()
        )
        self.register_builder(
            AL.Paradigm.LABEL_PROBABILITY_BASED_ENSEMBLE,
            LabelProbilityBasedEnsembleBuilder(),
        )
        self.register_builder(
            AL.Paradigm.LABEL_MIN_PROB_ENSEMBLE, LabelMinProbilityBasedEnsembleBuilder()
        )
        self.register_builder(AL.CustomMethods.AUTOTAR, AutoTARBuilder())
        self.register_builder(AL.CustomMethods.AUTOSTOP, AutoSTOPBuilder())
        self.register_constructor(AL.QueryType.RANDOM_SAMPLING, RandomSampling.builder)
        self.register_constructor(AL.QueryType.LEAST_CONFIDENCE, LeastConfidence)
        self.register_constructor(AL.QueryType.MAX_ENTROPY, EntropySampling)
        self.register_constructor(AL.QueryType.MARGIN_SAMPLING, MarginSampling)
        self.register_constructor(
            AL.QueryType.NEAR_DECISION_BOUNDARY, NearDecisionBoundary
        )
        self.register_constructor(AL.QueryType.LABELMAXIMIZER, LabelMaximizer)
        self.register_constructor(AL.QueryType.LABELUNCERTAINTY, LabelUncertainty)
        self.register_constructor(AL.QueryType.MOST_CERTAIN, MostCertainSampling)
        self.register_constructor(AL.QueryType.MOST_CONFIDENCE, MostConfidence)
        self.register_constructor(AL.QueryType.LABELMAXIMIZER_NEW, LabelMaximizerNew)
        self.register_constructor(
            AL.QueryType.LABELUNCERTAINTY_NEW, LabelUncertaintyNew
        )
        self.register_constructor(AL.QueryType.RANDOM_ML, RandomMLStrategy)
  



from os import PathLike
from typing import Any, Generic, Iterable, Optional, Sequence, Tuple

from instancelib.labels.encoder import DictionaryEncoder
import instancelib as il
import numpy as np
import numpy.typing as npt

from allib.typehints.typevars import IT, KT, LT
from .tsvm_sklearn import SKTSVM


class TSVMEncoder(DictionaryEncoder[LT]):
    def encode(self, labels: Iterable[LT]) -> npt.NDArray[Any]:
        label_list = list(labels)
        if not label_list:
            return np.array([-1])
        return super().encode(label_list)

    def initialize(self, labels: Iterable[LT]) -> None:
        return super().initialize(labels)


class TSVM(il.SkLearnVectorClassifier[IT, KT, LT], Generic[IT, KT, LT]):
    def __init__(
        self,
        storage_location: "Optional[PathLike[str]]" = None,
        filename: "Optional[PathLike[str]]" = None,
    ) -> None:
        estimator = SKTSVM()
        encoder = TSVMEncoder.from_list([])
        super().__init__(estimator, encoder, storage_location, filename)



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Tuple, Optional

import numpy as np
import numpy.typing as npt

from ..environment import AbstractEnvironment

class BaseBalancer(ABC):
    """Abstract class for balance strategies."""
    _environment: Optional[AbstractEnvironment]
    name = "BaseBalancer"
    
    def __call__(self, environment: AbstractEnvironment) -> BaseBalancer:
        """Attach environment to Balancer. This
        can be used to add additional data to the training set
        
        Parameters
        ----------
        environment : AbstractEnvironment
            The environment used for this project
        
        Returns
        -------
        BaseBalancer
            The Balancer with the attached Environment
        """        
        self._environment = environment
        return self

    @abstractmethod
    def resample(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        """Resample the training data
        
        Parameters
        ----------
        x_data : npt.NDArray[Any]
            The feature matrix of the training data
        y_data : npt.NDArray[Any]
            The encoded labels for all the training data
        
        Returns
        -------
        Tuple[npt.NDArray[Any], npt.NDArray[Any]]
            the resampled feature matrix, and corresponding labels.
        
       
        """        
        raise NotImplementedError

class IdentityBalancer(BaseBalancer):
    def __init__(self) -> None:
        self._environment = None
        
    def resample(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        """Resample the training data (identity function)
        
        Parameters
        ----------
        x_data : npt.NDArray[Any]
            The feature matrix of the training data
        y_data : npt.NDArray[Any]
            The encoded labels for all the training data
        
        Returns
        -------
        Tuple[npt.NDArray[Any], npt.NDArray[Any]]
            the resampled feature matrix, and corresponding labels.
        """        
        return x_data, y_data


from __future__ import annotations

import collections
import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import (Any, Deque, Dict, FrozenSet, Generic, Iterable, Optional,
                    Sequence, Set, Tuple, TypeVar, Union)

import pandas as pd  # type: ignore
from instancelib import Instance
from instancelib.labels import LabelProvider
from instancelib.labels.memory import MemoryLabelProvider
from instancelib.utils.to_key import to_key
from instancelib.typehints import KT, LT

from .base import BaseLogger

MT = TypeVar("MT")
ST = TypeVar("ST")

SampleMethod = Tuple[ST, Optional[LT]]

class Event(ABC, Generic[KT, LT, ST]):
    timestamp: datetime
    name: str = "Event"

    def __init__(self):
        self.taboo = ["timestamp", "name", "__orig_class__", "taboo"]
        self.timestamp = datetime.now()
    
    def __str__(self) -> str:
        def kvpair(x: Any, y: Any) -> str:
            return f"{x} => {y}"
        dict_filter = [(k, v) for (k, v) in self.__dict__.items() if k not in self.taboo]
        dict_tuples = itertools.starmap(kvpair, dict_filter)
        dict_str = ", ".join(dict_tuples)
        return f"{self.timestamp} :: {self.name} :: {dict_str}"
    
    def __repr__(self) -> str:
        return str(self)


class SampleEvent(Event[KT, LT, ST], Generic[KT, LT, ST]):
    name = "Sampling"

    def __init__(self, key: KT, method: SampleMethod[ST, LT]):
        super(SampleEvent, self).__init__()
        self.key = key
        self.method = method[0]
        self.method_label = method[1]
        


class LabelEvent(SampleEvent[KT, LT, ST], Generic[KT, LT, ST]):
    name = "Label"
    def __init__(self, key: KT, method: SampleMethod[ST, LT], *labels: LT):
        super().__init__(key, method)
        self.labels = frozenset(labels)
        self.taboo.append("labels")

    def __str__(self):
        label_str = ", ".join(map(str, self.labels))
        super_str = super(LabelEvent, self).__str__()
        return f"{super_str}, labels => [{label_str}]"

    def __repr__(self):
        return str(self)

@dataclass
class Snapshot(Generic[KT, LT]):
    ordering: Sequence[KT]
    metrics: Sequence[float]
    labeled: FrozenSet[KT]
    labels: LabelProvider[KT, LT]

class MemoryLogger(BaseLogger[KT, LT, SampleMethod[ST, LT]], Generic[KT, LT,ST]):
    def __init__(self, label_provider: LabelProvider[KT, LT]):
        self._labelset = label_provider.labelset

        self.sample_dict: Dict[KT, Set[SampleMethod[ST, LT]]] = dict()
        self.sample_dict_inv: Dict[SampleMethod[ST, LT], Set[KT]] = dict()
        
        self.snapshots: Deque[Snapshot[KT, LT]] = collections.deque()
        self.sample_history: Deque[SampleEvent[KT, LT, ST]] = collections.deque()
        self.event_history: Deque[Event[KT, LT, ST]] = collections.deque()
        self.label_history: Deque[LabelEvent[KT, LT, ST]] = collections.deque()
        self.label_order: Dict[KT, int] = dict()
        self.label_it = 0

    @property
    def labelset(self) -> FrozenSet[LT]: 
        return self._labelset
        
    def log_ordering(self, ordering: Sequence[KT], metrics: Sequence[float], labeled: Iterable[KT], labels: LabelProvider[KT, LT]):
        labeled = frozenset(labeled)
        labels_copy = MemoryLabelProvider[KT, LT].from_provider(labels)
        self.snapshots.append(Snapshot[KT, LT](ordering, metrics, labeled, labels_copy))

    def log_sample(self, x: Union[KT, Instance[KT, Any, Any, Any]], sample_method: SampleMethod[ST, LT]) -> None:
        key = to_key(x)
        self.sample_dict.setdefault(key, set()).add(sample_method)
        self.sample_dict_inv.setdefault(sample_method, set()).add(key)
        event = SampleEvent[KT, LT, ST](key, sample_method)
        self.sample_history.append(event)
        self.event_history.append(event)

    def log_label(self, x: Union[KT, Instance[KT, Any, Any, Any]], sample_method: SampleMethod[ST, LT], *labels: LT):
        key = to_key(x)
        event = LabelEvent[KT, LT, ST](key, sample_method, *labels)
        self.event_history.append(event)
        self.label_history.append(event)
        self.label_order[key] = self.label_it
        self.label_it = self.label_it + 1

    def get_label_order(self, x: Union[KT, Instance[KT, Any, Any, Any]]) -> int:
        key = to_key(x)
        return self.label_order[key]
    
    def get_sampled_info(self, x: Union[KT, Instance[KT, Any, Any, Any]]):
        key = to_key(x)
        return frozenset(self.sample_dict.setdefault(key, set()))

    def get_instances_by_method(self, sample_method: SampleMethod[ST, LT]):
        return frozenset(self.sample_dict_inv.setdefault(sample_method, set()))

    def get_label_table(self) -> pd.DataFrame:
        def row_generator():
            for event in self.label_history:
                doc_labels = event.labels
                label_dict = {
                    label: (label in doc_labels) for label in self.labelset}
                event_dict = {
                    "timestamp": event.timestamp,
                    "instance_id": event.key,
                    "method": event.method,
                    "method_label": event.method_label,
                }
                yield {**event_dict, **label_dict}
        dataframe = pd.DataFrame(list(row_generator()))
        return dataframe

    def get_label_cumsum_table(self) -> pd.DataFrame:
        def row_generator():
            label_sum_dict: Dict[Tuple[LT, ST, Optional[LT]], int] = dict()
            for event in self.label_history:
                for label in event.labels:
                    dict_key = (label, event.method, event.method_label)
                    label_sum_dict.setdefault(dict_key, 0)
                    label_sum_dict[dict_key] += 1
                event_dict = {
                    "timestamp": event.timestamp,
                    "instance_id": event.key,
                }
                yield {**event_dict, **label_sum_dict}
        dataframe = pd.DataFrame(list(row_generator()))
        return dataframe



from __future__ import annotations

from abc import ABC
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Iterator,
    MutableMapping,
    Sequence,
    Union,
)
from uuid import UUID

import instancelib as il

from instancelib import InstanceProvider
from instancelib.instances.memory import (
    DataPoint,
    DataPointProvider,
    MemoryBucketProvider,
)
from instancelib.labels.base import LabelProvider
from instancelib.labels.memory import MemoryLabelProvider
from instancelib.typehints import DT, KT, LT, RT, VT
from instancelib.utils.func import union

from ..history import MemoryLogger
from ..history.base import BaseLogger
from .base import IT, AbstractEnvironment

# TODO Adjust MemoryEnvironment Generic Type (ADD ST)


class AbstractMemoryEnvironment(
    AbstractEnvironment[IT, KT, DT, VT, RT, LT], ABC, Generic[IT, KT, DT, VT, RT, LT]
):

    _public_dataset: InstanceProvider[IT, KT, DT, VT, RT]
    _dataset: InstanceProvider[IT, KT, DT, VT, RT]
    _unlabeled: InstanceProvider[IT, KT, DT, VT, RT]
    _labeled: InstanceProvider[IT, KT, DT, VT, RT]
    _labelprovider: LabelProvider[KT, LT]
    _truth: LabelProvider[KT, LT]
    _logger: BaseLogger[KT, LT, Any]
    _named_providers: MutableMapping[str, InstanceProvider[IT, KT, DT, VT, RT]]

    def __contains__(self, __o: object) -> bool:
        return __o in self._named_providers

    def __getitem__(self, __k: str) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self._named_providers[__k]

    def __setitem__(self, __k: str, __v: InstanceProvider[IT, KT, DT, VT, RT]) -> None:
        self.set_named_provider(__k, __v)

    def __len__(self) -> int:
        return len(self._named_providers)

    def __delitem__(self, __v: str) -> None:
        del self._named_providers[__v]

    def __iter__(self) -> Iterator[str]:
        return iter(self._named_providers)

    @property
    def dataset(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self._public_dataset

    @property
    def all_instances(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self._dataset

    @property
    def labels(self) -> LabelProvider[KT, LT]:
        return self._labelprovider

    @property
    def logger(self) -> BaseLogger[KT, LT, Any]:
        return self._logger

    @property
    def unlabeled(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self._unlabeled

    @property
    def labeled(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self._labeled

    @property
    def truth(self) -> LabelProvider[KT, LT]:
        return self._truth

    def create_bucket(self, keys: Iterable[KT]) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return MemoryBucketProvider[IT, KT, DT, VT, RT](self._dataset, keys)

    def create_empty_provider(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        return self.create_bucket([])

    def set_named_provider(
        self, name: str, value: InstanceProvider[IT, KT, DT, VT, RT]
    ):
        self._named_providers[name] = value

    def create_named_provider(
        self, name: str, keys: Iterable[KT] = list()
    ) -> InstanceProvider[IT, KT, DT, VT, RT]:
        self._named_providers[name] = self.create_bucket(keys)
        return self._named_providers[name]


class MemoryEnvironment(
    AbstractMemoryEnvironment[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    def __init__(
        self,
        dataset: InstanceProvider[IT, KT, DT, VT, RT],
        unlabeled: InstanceProvider[IT, KT, DT, VT, RT],
        labeled: InstanceProvider[IT, KT, DT, VT, RT],
        named_providers: MutableMapping[str, InstanceProvider[IT, KT, DT, VT, RT]],
        public_dataset: InstanceProvider[IT, KT, DT, VT, RT],
        labelprovider: LabelProvider[KT, LT],
        logger: BaseLogger[KT, LT, Any],
        truth: LabelProvider[KT, LT],
    ):
        self._dataset = dataset
        self._unlabeled = unlabeled
        self._labeled = labeled
        self._labelprovider = labelprovider
        self._named_providers = named_providers
        self._logger = logger
        self._truth = truth
        self._public_dataset = public_dataset

    @classmethod
    def from_environment(
        cls,
        environment: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        shared_labels: bool = True,
        *args,
        **kwargs,
    ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        dataset = environment.all_instances
        unlabeled = MemoryBucketProvider(dataset, environment.unlabeled.key_list)
        labeled = MemoryBucketProvider(dataset, environment.labeled.key_list)
        if shared_labels:
            labels = environment.labels
        else:
            labels = MemoryLabelProvider[KT, LT].from_data(
                environment.labels.labelset, [], []
            )
        logger = environment.logger
        truth = environment.truth
        named_providers: Dict[str, InstanceProvider[IT, KT, DT, VT, RT]] = {
            key: MemoryBucketProvider(dataset, prov.key_list)
            for key, prov in environment.items()
        }
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )

    @classmethod
    def from_environment_only_data(
        cls, environment: AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        dataset = environment.all_instances
        unlabeled = MemoryBucketProvider(dataset, dataset.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[KT, LT](environment.labels.labelset, {}, {})
        logger = MemoryLogger[KT, LT, Any](labels)
        truth = environment.truth
        named_providers: Dict[str, InstanceProvider[IT, KT, DT, VT, RT]] = {
            key: MemoryBucketProvider(dataset, prov.key_list)
            for key, prov in environment.items()
        }
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )

    @classmethod
    def from_instancelib(
        cls, environment: il.AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        dataset = environment.all_instances
        labeled_docs = union(
            *(
                environment.labels.get_instances_by_label(label)
                for label in environment.labels.labelset
            )
        )
        unlabeled_docs = frozenset(dataset.key_list).difference(labeled_docs)
        unlabeled = MemoryBucketProvider(dataset, unlabeled_docs)
        labeled = MemoryBucketProvider(dataset, labeled_docs)
        labels = MemoryLabelProvider[KT, LT].from_provider(environment.labels)
        environment.labels
        logger = MemoryLogger[KT, LT, Any](labels)
        truth = MemoryLabelProvider[KT, LT].from_provider(environment.labels)
        named_providers: Dict[str, InstanceProvider[IT, KT, DT, VT, RT]] = {
            key: MemoryBucketProvider(dataset, prov.key_list)
            for key, prov in environment.items()
        }
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )

    @classmethod
    def from_instancelib_simulation(
        cls, environment: il.AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        dataset = environment.all_instances
        unlabeled = MemoryBucketProvider(dataset, dataset.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[KT, LT].from_data(
            environment.labels.labelset, [], []
        )
        logger = MemoryLogger[KT, LT, Any](labels)
        truth = MemoryLabelProvider[KT, LT].from_provider(environment.labels)
        named_providers: Dict[str, InstanceProvider[IT, KT, DT, VT, RT]] = {
            key: MemoryBucketProvider(dataset, prov.key_list)
            for key, prov in environment.items()
        }
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )

    @classmethod
    def from_instancelib_simulation_heldout(
        cls,
        environment: il.AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        train_set: il.InstanceProvider[IT, KT, DT, VT, RT],
    ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        dataset = environment.all_instances
        unlabeled = MemoryBucketProvider(dataset, train_set.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[KT, LT].from_data(
            environment.labels.labelset, [], []
        )
        logger = MemoryLogger[KT, LT, Any](labels)
        truth = MemoryLabelProvider[KT, LT].from_provider(environment.labels)
        named_providers: Dict[str, InstanceProvider[IT, KT, DT, VT, RT]] = {
            key: MemoryBucketProvider(dataset, prov.key_list)
            for key, prov in environment.items()
        }
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )


class DataPointEnvironment(
    MemoryEnvironment[
        DataPoint[Union[KT, UUID], DT, VT, RT], Union[KT, UUID], DT, VT, RT, LT
    ],
    Generic[KT, DT, VT, RT, LT],
):
    @classmethod
    def from_data(
        cls,
        target_labels: Iterable[LT],
        indices: Sequence[KT],
        data: Sequence[DT],
        ground_truth: Sequence[Iterable[LT]],
        vectors: Sequence[VT],
    ) -> DataPointEnvironment[KT, DT, VT, RT, LT]:
        dataset = DataPointProvider[KT, DT, VT, RT].from_data_and_indices(
            indices, data, vectors
        )
        unlabeled = MemoryBucketProvider(dataset, dataset.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[Union[KT, UUID], LT].from_data(
            target_labels, indices, []
        )
        logger = MemoryLogger[Union[KT, UUID], LT, Any](labels)
        truth = MemoryLabelProvider[Union[KT, UUID], LT].from_data(
            target_labels, indices, ground_truth
        )
        named_providers = dict()
        public_dataset = MemoryBucketProvider(dataset, dataset.key_list)
        return cls(
            dataset,
            unlabeled,
            labeled,
            named_providers,
            public_dataset,
            labels,
            logger,
            truth,
        )



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
from typing import Any


class AbstractSelectionCriterion(ABC):
    name: str = "AbstractSelectionCriterion"

    def __init__(self, *_, **__) -> None:
        pass

    @abstractmethod
    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        """Calculates the selection metric given a probability matrix

        Parameters
        ----------
        prob_mat : npt.NDArray[Any]
            The probability matrix with rows of class probabilities. 
            Shape should be ``(n_instances, n_classes)``

        Returns
        -------
        npt.NDArray[Any]
            The result of the selection metrix. This has as shape
            ``(n_instances, )``

        """
        raise NotImplementedError


# import math
# import random
# from abc import ABC, abstractmethod
# from typing import (Any, AnyStr, Callable, Dict, Iterator, List, Optional, Set,
#                     Tuple)

# import numpy as np
# import numpy.typing as npt
# import pandas as pd
# from numpy.linalg import inv
# from numpy.random import multivariate_normal
# from sklearn.base import BaseEstimator
# from sklearn.cluster import KMeans, MiniBatchKMeans
# from sklearn.metrics.pairwise import cosine_similarity
# from scipy import spatial

# from .base import AbstractActiveLearner
# from .random import RandomSampling
# from .datapoint import ClusteredDataPoint
# from .oracles import OracleFunction

# from ..instances import Instance, InstanceProvider


# class ATSCluster(RandomSampling):
#     identifier: int

#     mean_distance: float
#     var_distance: float
#     super_size: int

#     def __init__(
#             self,
#             provider: InstanceProvider,
#             labels: List[str],
#             identifier: int,
#             datapoints: List[ClusteredDataPoint],
#             super_size: int,
#             estimator: BaseEstimator
#     ) -> None:
#         super().__init__(provider, labels, datapoints, estimator)
#         self.id = identifier

#         vec_mat = np.matrix(np.vstack([d.vector for d in datapoints]))
#         self.mean_distance = np.average(cosine_similarity(vec_mat))
#         self.var_distance = np.var(cosine_similarity(vec_mat))
#         self.super_size = super_size

#     def get_context_vector(self) -> npt.NDArray[Any]:
#         """Generates a context vector that describes the cluster

#         Returns
#         -------
#         npt.NDArray[Any]
#             A (normalized) feature vector of size `4 + len(labels)`
#         """
#         vec = [
#             self.mean_distance,
#             self.var_distance,
#             self.size / self.super_size,  # TODO Check if it is better to not scale the size
#             self.get_labeled_ratio()
#         ]
#         for label in self.labels:
#             ratio = len(self.label_dict[label]) / self.size
#             vec.append(ratio)
#         npvec = np.array(vec)
#         return npvec / np.linalg.norm(npvec)


# class ActiveThompsonSampling(AbstractActiveLearner):
#     number_of_clusters: int
#     labels: List[str]

#     cluster_dict: Dict[int, ATSCluster]
#     cluster_dict_doc: Dict[str, ATSCluster]
#     dims: int
#     mat_b: npt.NDArray[Any]
#     vec_mu_hat: npt.NDArray[Any]
#     vec_f: npt.NDArray[Any]
#     v_squared: float = 0.25
#     t: int

#     rewards: List[float]

#     alpha = 0.61
#     beta = 0.12
#     stored_hypothesis: npt.NDArray[Any]

#     def __init__(
#             self,
#             provider: InstanceProvider,
#             labels: List[str],
#             datapoints: List[ClusteredDataPoint],
#             estimator: BaseEstimator) -> None:
#         super().__init__(provider, labels, datapoints, estimator)

#         cluster_ids = {d.cluster for d in datapoints}
#         cluster_dict = {i: [] for i in cluster_ids}

#         for d in datapoints:
#             cluster_dict[d.cluster].append(d)
#         self.cluster_dict = {
#             i: ATSCluster(provider, labels, i, datapoints, len(datapoints), estimator)
#             for i, datapoints in cluster_dict.items()
#         }
#         self.cluster_dict_doc = {
#             d.index: self.cluster_dict[d.cluster] for d in datapoints}

#         self.t = 0
#         self.dims = 4 + len(labels)

#         self.mat_b = np.identity(self.dims)
#         self.vec_mu_hat = np.zeros(shape=(self.dims))
#         self.vec_f = np.zeros(shape=(self.dims))
#         self.rewards = []
#         self.stored_hypothesis = None

#     def label_instance(self, doc_id: str, label: str) -> None:
#         """Labels the instance and discard it from the unlabeled set

#         Parameters
#         ----------
#         doc_id : str
#             The identifier of the document
#         label : str
#             The label of the document
#         """
#         self.label_dict[label].append(doc_id)
#         self.label_dict_inv[doc_id].append(label)
#         self.unlabeled.discard(doc_id)
#         self.labeled.add(doc_id)

#         # Update the cluster to which the document belongs
#         cluster = self.cluster_dict_doc[doc_id]
#         cluster.label_instance(doc_id, label)

#     def calculate_hypothesis(self) -> npt.NDArray[Any]:
#         """Calculate the hypothesis, the result of the prediction of all documents in all clusters (flattened)

#         Returns
#         -------
#         npt.NDArray[Any]
#             The resulting flattened vector
#         """
#         vectors = np.array([vec for vec in self.vector_generator()])
#         pred = self.estimator.predict_proba(vectors)
#         return pred.toarray().flatten()

#     def store_hypothesis(self, hypothesis: npt.NDArray[Any]) -> None:
#         """Store the hypothesis (overwrite)

#         Parameters
#         ----------
#         hypothesis : npt.NDArray[Any]
#             The (new) hypothesis
#         """
#         self.stored_hypothesis = hypothesis

#     def decreasing_function(self) -> float:
#         """The decreasing function that scales the rewards 
#            TODO: Check if this is useful or harmful        
#         Returns
#         -------
#         float
#             Decreasing function
#         """
#         return self.alpha * math.exp(-self.beta * self.t)

#     def calculate_reward(self, old_hypothesis, new_hypothesis) -> float:
#         """Calculate the reward, which is the amount in which the old hypothesis differs from the old

#         Parameters
#         ----------
#         old_hypothesis : [type]
#             Old hypothesis
#         new_hypothesis : [type]
#             New hypothesis

#         Returns
#         -------
#         float
#             Reward
#         """
#         cos_sim = 1.0
#         if old_hypothesis is not None:
#             cos_sim = 1 - \
#                 spatial.distance.cosine(old_hypothesis, new_hypothesis)
#         # $y_t = 2 \cdot \frac{\cos^-1(d(h, h_t-1))}{\pi}$
#         yt = 2 * math.acos(cos_sim) / math.pi
#         rt = yt  # / self.decreasing_function()
#         return rt

#     def select_random_cluster(self, non_empty=False) -> ATSCluster:
#         """Select a random cluster (for example, during initialization)

#         Parameters
#         ----------
#         non_empty : bool, optional
#             If we only desire to select clusters that are not empty, by default False

#         Returns
#         -------
#         ATSCluster
#             A randomly selected cluster
#         """
#         if non_empty:
#             cluster_ids = [cluster_id for cluster_id,
#                            c in self.cluster_dict.items() if c.len_unlabeled() > 0]
#         else:
#             cluster_ids = [cluster_id for cluster_id in self.cluster_dict]
#         random_id = random.choice(cluster_ids)
#         return self.cluster_dict[random_id]

#     def select_cluster(self) -> ATSCluster:
#         """Select most optimal Cluster, according to the Active Thompson Sampling algorithm

#         Returns
#         -------
#         ATSCluster
#             The estimated most optimal cluster 
#         """
#         vec_mu_sample = np.random.multivariate_normal(
#             self.vec_mu_hat, self.v_squared * inv(self.mat_b))
#         con_vecs = [(i, c.get_context_vector())
#                     for i, c in self.cluster_dict.items() if c.len_unlabeled() > 0]
#         con_vecs = [(i, np.matmul(np.transpose(c), vec_mu_sample))
#                     for i, c in con_vecs]
#         cluster_id, _ = max(con_vecs, key=lambda v: v[1])
#         return self.cluster_dict[cluster_id]

#     def iterate_al(self, f_oracle: OracleFunction) -> None:
#         """Perform one iteration of the active learning cycle

#         Parameters
#         ----------
#         f_oracle : OracleFunction
#             Oracle function
#         """
#         # Select cluster and sample data
#         cluster = self.select_cluster()
#         context_vector = cluster.get_context_vector()
#         doc_id, vec, data = cluster.sample()

#         # Query the oracle and label the instance
#         label = f_oracle(doc_id, vec, data, self.labels)
#         self.label_instance(doc_id, label)

#         # Calculate the current hypothesis or retrieve it
#         if self.stored_hypothesis is None:
#             old_hypothesis = self.calculate_hypothesis()
#         else:
#             old_hypothesis = self.stored_hypothesis

#         # Retrain the Machine Learning Model
#         self.fit()

#         # Calculate changes in the predictions and calculate the reward
#         new_hypothesis = self.calculate_hypothesis()
#         reward = self.calculate_reward(old_hypothesis, new_hypothesis)

#         # Store the hypothesis and reward
#         self.rewards.append(reward)
#         self.store_hypothesis(new_hypothesis)

#         # Update sample parameters
#         update_b = context_vector * context_vector.reshape(-1, 1)
#         self.mat_b = self.mat_b + update_b
#         self.vec_f = self.vec_f + context_vector * reward
#         self.vec_mu_hat = np.matmul(inv(self.mat_b), self.vec_f)
#         self.t = self.t+1

#     def initialize(self, f_oracle: OracleFunction, it_max) -> None:
#         """Label `it_max` randomly selected documents 

#         Parameters
#         ----------
#         f_oracle :
#             Oracle function 
#         it_max : 
#             The number of iterations / documents
#         """
#         for _ in range(it_max):
#             doc_id, vec, data = self.sample()
#             label = f_oracle(doc_id, vec, data, self.labels)
#             self.label_instance(doc_id, label)

#     def initial_fit(self) -> None:
#         """Store initial hypothesis
#         """
#         self.fit()
#         self.stored_hypothesis = self.calculate_hypothesis()



from os import PathLike
from pathlib import Path
from typing import Union


def create_dir_if_not_exists(path: Union[str, Path]) -> None:
    path = Path(path)
    if not path.exists():
       path.mkdir(parents=True)


from enum import Enum

class Component(Enum):
    FEATURE_EXTRACTION = "FeatureExtraction"
    CLASSIFIER = "Classifier"
    ACTIVELEARNER = "ActiveLearner"
    FALLBACK = "Fallback"
    ENVIRONMENT = "Environment"
    VECTORIZER = "Vectorizer"
    BALANCER = "Balancer"
    SELECTION_CRITERION = "SelectionCriterion"


import random as rnd

import numpy
import sklearn.metrics
from sklearn.base import ClassifierMixin
from sklearn.linear_model import LogisticRegression as LR

from .qns3vm import QN_S3VM


class SKTSVM(ClassifierMixin):
    """
    Scikit-learn wrapper for transductive SVM (SKTSVM)

    Wraps QN-S3VM by Fabian Gieseke, Antti Airola, Tapio Pahikkala, Oliver Kramer (see http://www.fabiangieseke.de/index.php/code/qns3vm)
    as a scikit-learn BaseEstimator, and provides probability estimates using Platt scaling
    Parameters
    ----------
    C : float, optional (default=1.0)
        Penalty parameter C of the error term.
    kernel : string, optional (default='rbf')
         Specifies the kernel type to be used in the algorithm.
         It must be 'linear' or 'rbf'
    gamma : float, optional (default=0.0)
        Kernel coefficient for 'rbf'
    lamU: float, optional (default=1.0)
        cost parameter that determines influence of unlabeled patterns
        must be float >0
    probability: boolean, optional (default=False)
        Whether to enable probability estimates. This must be enabled prior
        to calling `fit`, and will slow down that method.
    """

    # lamU -- cost parameter that determines influence of unlabeled patterns (default 1, must be float > 0)
    def __init__(self, kernel="Linear", C=1e-4, gamma=0.5, lamU=1.0, probability=True):
        self.random_generator = rnd.Random()
        self.kernel = kernel
        self.C = C
        self.gamma = gamma
        self.lamU = lamU
        self.probability = probability

    def fit(self, X, y):  # -1 for unlabeled
        """Fit the model according to the given training data.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target vector relative to X
            Must be 0 or 1 for labeled and -1 for unlabeled instances
        Returns
        -------
        self : object
            Returns self.
        """

        # http://www.fabiangieseke.de/index.php/code/qns3vm
        y = y.reshape(-1)

        unlabeledX = X[y == -1, :].tolist()
        labeledX = X[y != -1, :].tolist()
        labeledy = y[y != -1]

        # convert class 0 to -1 for tsvm
        labeledy[labeledy == 0] = -1
        labeledy = labeledy.tolist()

        if "rbf" in self.kernel.lower():
            self.model = QN_S3VM(
                labeledX,
                labeledy,
                unlabeledX,
                self.random_generator,
                lam=self.C,
                lamU=self.lamU,
                kernel_type="RBF",
                sigma=self.gamma,
            )
        else:
            self.model = QN_S3VM(
                labeledX,
                labeledy,
                unlabeledX,
                self.random_generator,
                lam=self.C,
                lamU=self.lamU,
            )

        self.model.train()

        # probabilities by Platt scaling
        if self.probability:
            self.plattlr = LR(solver="lbfgs")
            preds = self.model.mygetPreds(labeledX)
            self.plattlr.fit(preds.reshape(-1, 1), labeledy)

    def predict_proba(self, X):
        """Compute probabilities of possible outcomes for samples in X.
        The model need to have probability information computed at training
        time: fit with attribute `probability` set to True.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        Returns
        -------
        T : array-like, shape = [n_samples, n_classes]
            Returns the probability of the sample for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute `classes_`.
        """

        if self.probability:
            preds = self.model.mygetPreds(X.tolist())
            return self.plattlr.predict_proba(preds.reshape(-1, 1))
        else:
            raise RuntimeError(
                "Probabilities were not calculated for this model - make sure you pass probability=True to the constructor"
            )

    def predict(self, X):
        """Perform classification on samples in X.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        Returns
        -------
        y_pred : array, shape = [n_samples]
            Class labels for samples in X.
        """

        y = numpy.array(self.model.getPredictions(X.tolist()))
        y[y == -1] = 0
        return y

    def score(self, X, y, sample_weight=None):
        return sklearn.metrics.accuracy_score(
            y, self.predict(X), sample_weight=sample_weight
        )



from __future__ import annotations

from typing import Sequence, Any

import numpy as np
import numpy.typing as npt

from instancelib import Instance

from .base import BaseVectorizer

InstanceList = Sequence[Instance[Any, str, npt.NDArray[Any], Any]] # type: ignore

class TextInstanceVectorizer(BaseVectorizer[Instance[Any, str, npt.NDArray[Any], Any]]): 
    _name = "TextInstanceVectorizer"
    def __init__(self,
                 vectorizer: BaseVectorizer[str],
                 ) -> None:
        super().__init__()
        self.innermodel = vectorizer

    def fitted(self) -> bool:
        return self.innermodel.fitted

    def fit(self, x_data: InstanceList, **kwargs: Any) -> TextInstanceVectorizer:
        texts = [x.data for x in x_data]
        self.innermodel.fit(texts)
        return self

    def transform(self, x_data: InstanceList, **kwargs: Any) -> npt.NDArray[Any][Any]:
        texts = [x.data for x in x_data]
        return self.innermodel.transform(texts) # type: ignore

    def fit_transform(self, x_data: InstanceList, **kwargs: Any) -> npt.NDArray[Any][Any]:
        texts = [x.data for x in x_data]
        return self.innermodel.fit_transform(texts) # type: ignore



from pathlib import Path
from allib import app
import argparse
from multiprocessing import Pool

from allib.configurations.catalog import ExperimentCombination

from .module.catalog import ActiveLearningTasks

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="allib",
        description="Active Learning Library (allib) - Benchmarking tool"
    )
    parser.add_argument("-d", "--dataset", help="The path to the dataset", type=Path)
    parser.add_argument("-t", "--target", help="The target of the results", type=Path)
    parser.add_argument("-e", "--exp_choice", help="The experiment method", type=ExperimentCombination)
    parser.add_argument("--problem", "--task", metavar="problem", default="TAR", type=ActiveLearningTasks, help="The problem specification")
    parser.add_argument("-p", "--pos_label", metavar="POS", default="Relevant", help="The label that denotes the positive class")
    parser.add_argument("-n", "--neg_label", metavar="NEG", default="Irrelevant", help="The label that denotes the negative class")
    parser.add_argument("-q", "--nthreads", default=4, type=int)
    parser.add_argument("-m", "--samplesize", default=15, type=int)
    args = parser.parse_args()
    pool = Pool(args.nthreads)
    if args.problem == ActiveLearningTasks.TAR:
        app.tar_benchmark(args.dataset, 
                          args.target, 
                          args.exp_choice, 
                          args.pos_label,
                          args.neg_label)
    else:
        raise NotImplementedError("Other tasks have no Benchmark procedure yet")


from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    MutableMapping,
    Sequence,
    Union,
)

import instancelib as il
from instancelib import InstanceProvider, LabelProvider
from instancelib.analysis.base import MulticlassModelMetrics
from instancelib.utils.func import value_map

from ..activelearning.base import ActiveLearner
from ..activelearning.insclass import ILMLBased
from ..estimation.base import AbstractEstimator, Estimate
from ..stopcriterion.base import AbstractStopCriterion
from ..typehints import DT, IT, KT, LT, RT, VT
from .statistics import ALStats, LabelALStatistics


class ExperimentIterator(Generic[IT, KT, DT, VT, RT, LT]):
    learner: ActiveLearner[IT, KT, DT, VT, RT, LT]
    it: int
    batch_size: int
    stop_interval: Mapping[str, int]
    stopping_criteria: Mapping[str, AbstractStopCriterion[LT]]
    estimators: Mapping[str, AbstractEstimator[IT, KT, DT, VT, RT, LT]]

    def __init__(
        self,
        learner: ActiveLearner[IT, KT, DT, VT, RT, LT],
        pos_label: LT,
        neg_label: LT,
        stopping_criteria: Mapping[str, AbstractStopCriterion[LT]],
        estimators: Mapping[str, AbstractEstimator[IT, KT, DT, VT, RT, LT]],
        batch_size: int = 1,
        stop_interval: Union[int, Mapping[str, int]] = 1,
        estimation_interval: Union[int, Mapping[str, int]] = 1,
        iteration_hooks: Sequence[
            Callable[[ActiveLearner[IT, KT, DT, VT, RT, LT]], Any]
        ] = list(),
        estimator_hooks: Mapping[
            str, Callable[[AbstractEstimator[IT, KT, DT, VT, RT, LT]], Any]
        ] = dict(),
    ) -> None:
        # Iteration tracker
        self.it = 0
        # Algorithm selection
        self.learner = learner
        self.stopping_criteria = stopping_criteria
        self.estimators = estimators

        # Labels
        self.pos_label = pos_label
        self.neg_label = neg_label

        # Estimation tracker
        self.estimation_tracker: Dict[str, Estimate] = dict()
        self.iteration_hooks = iteration_hooks
        self.estimator_hooks = estimator_hooks

        # Batch sizes
        self.batch_size = batch_size
        self.stop_interval = (
            {k: stop_interval for k in self.stopping_criteria}
            if isinstance(stop_interval, int)
            else stop_interval
        )
        self.estimation_interval = (
            {k: estimation_interval for k in self.estimators}
            if isinstance(estimation_interval, int)
            else estimation_interval
        )

    def _retrain(self) -> None:
        if self.it % self.batch_size == 0:
            self.learner.update_ordering()

    def determine_stop(self) -> Mapping[str, bool]:
        result: Dict[str, bool] = dict()
        for k, crit in self.stopping_criteria.items():
            if self.it % self.stop_interval[k] == 0:
                crit.update(self.learner)
            result[k] = crit.stop_criterion
        return result

    def _estimate_recall(self) -> Mapping[str, Estimate]:
        for k, estimator in self.estimators.items():
            if self.it % self.estimation_interval[k] == 0:
                estimation = estimator(self.learner, self.pos_label)
                self.estimation_tracker[k] = estimation
        return self.estimation_tracker

    @property
    def finished(self) -> bool:
        return self.learner.env.unlabeled.empty

    @property
    def recall_estimate(self) -> Mapping[str, Estimate]:
        return self.estimation_tracker

    def _query_and_label(self) -> None:
        instance = next(self.learner)
        oracle_labels = self.learner.env.truth.get_labels(instance)

        # Set the labels in the active learner
        self.learner.env.labels.set_labels(instance, *oracle_labels)
        self.learner.set_as_labeled(instance)

    def call_hooks(self) -> Sequence[Any]:
        results = [hook(self.learner) for hook in self.iteration_hooks]
        return results

    def call_estimate_hooks(self) -> Mapping[str, Any]:
        results = {
            k: hook(self.estimators[k]) for k, hook in self.estimator_hooks.items()
        }
        return results

    def iter_increment(self) -> None:
        self.it += 1

    def __call__(self) -> Mapping[str, bool]:
        self._retrain()
        self._query_and_label()
        self._estimate_recall()
        stop_result = self.determine_stop()
        self.call_hooks()
        self.iter_increment()
        return stop_result


class ClassificationExperiment(Generic[IT, KT, DT, VT, RT, LT]):
    learners: Mapping[str, ActiveLearner[IT, KT, DT, VT, RT, LT]]
    it: int
    batch_size: int
    stop_interval: Mapping[str, int]
    test_set: InstanceProvider[IT, KT, DT, VT, RT]
    ground_truth: LabelProvider[KT, LT]
    label_stats_tracker: MutableMapping[
        int, Mapping[str, Mapping[LT, LabelALStatistics]]
    ]
    stopping_criteria: Mapping[str, Mapping[str, AbstractStopCriterion[LT]]]
    estimators: Mapping[str, AbstractEstimator[IT, KT, DT, VT, RT, LT]]
    iteration_hooks: Mapping[
        str, Sequence[Callable[[ActiveLearner[IT, KT, DT, VT, RT, LT]], Any]]
    ]
    main_stats_tracker: MutableMapping[int, Mapping[str, ALStats]]
    metrics_tracker: MutableMapping[int, Mapping[str, MulticlassModelMetrics]]
    stop_tracker: MutableMapping[int, Mapping[str, Mapping[str, bool]]]

    def __init__(
        self,
        learners: Mapping[str, ActiveLearner[IT, KT, DT, VT, RT, LT]],
        test_set: InstanceProvider[IT, KT, DT, VT, RT],
        ground_truth: LabelProvider[KT, LT],
        stopping_criteria: Mapping[
            str, Mapping[str, AbstractStopCriterion[LT]]
        ] = dict(),
        batch_size: int = 1,
        stop_interval: Union[int, Mapping[str, int]] = 1,
        iteration_hooks: Mapping[
            str, Sequence[Callable[[ActiveLearner[IT, KT, DT, VT, RT, LT]], Any]]
        ] = dict(),
    ) -> None:
        # Iteration tracker
        self.it = 0
        # Algorithm selection
        self.learners = learners
        self.test_set = test_set
        self.ground_truth = ground_truth
        self.stopping_criteria = stopping_criteria

        self.iteration_hooks = iteration_hooks

        self.label_stats_tracker = dict()
        self.metrics_tracker = dict()
        self.stop_tracker = dict()
        self.main_stats_tracker = dict()

        # Batch sizes
        self.batch_size = batch_size
        self.stop_interval = (
            {k: stop_interval for k in self.stopping_criteria}
            if isinstance(stop_interval, int)
            else stop_interval
        )

    def _retrain(self) -> None:
        if self.it % self.batch_size == 0:
            for learner in self.learners.values():
                learner.update_ordering()

    def determine_stop(self) -> Mapping[str, Mapping[str, bool]]:
        result: Dict[str, Dict[str, bool]] = dict()
        for l_key, learner in self.learners.items():
            if self.stopping_criteria and l_key in self.stopping_criteria:
                for k, crit in self.stopping_criteria[l_key].items():
                    if self.it % self.stop_interval[k] == 0:
                        crit.update(learner)
                    result[l_key][k] = crit.stop_criterion
        return result

    def _label_stats(self) -> Mapping[str, Mapping[LT, LabelALStatistics[LT]]]:
        result: Dict[str, Dict[LT, LabelALStatistics]] = dict()
        for l_key, learner in self.learners.items():
            for label in learner.env.labels.labelset:
                subset = learner.env.get_subset_by_labels(learner.env.labeled, label)
                generated = frozenset(
                    learner.env.get_subset_by_labels(learner.env.all_instances, label)
                ).difference(subset)
                stats = LabelALStatistics(label, len(subset), len(generated))
                result.setdefault(l_key, dict())[label] = stats
        return result

    def _al_stats(self) -> Mapping[str, ALStats]:
        result = {
            l_key: ALStats(
                len(learner.env.unlabeled),
                len(learner.env.labeled),
                len(learner.env.dataset),
            )
            for l_key, learner in self.learners.items()
        }
        return result

    @property
    def label_stats(self) -> Mapping[str, Mapping[LT, LabelALStatistics]]:
        return self.label_stats_tracker[self.it - 1]

    @property
    def stop_result(self) -> Mapping[str, Mapping[str, bool]]:
        return self.stop_tracker[self.it - 1]

    def _calculate_metrics(self) -> Mapping[str, MulticlassModelMetrics]:
        result: Dict[str, MulticlassModelMetrics] = dict()
        for l_key, learner in self.learners.items():
            if isinstance(learner, ILMLBased):
                result[l_key] = il.classifier_performance(
                    learner.classifier, self.test_set, self.ground_truth
                )
        return result

    @property
    def finished(self) -> bool:
        return all((learner.env.unlabeled.empty for learner in self.learners.values()))

    @property
    def metrics(self) -> Mapping[str, MulticlassModelMetrics]:
        return self.metrics_tracker[self.it - 1]

    @property
    def main_stats(self) -> Mapping[str, ALStats]:
        return self.main_stats_tracker[self.it - 1]

    def _query_and_label(self) -> None:
        for learner in self.learners.values():
            instance = next(learner)
            oracle_labels = learner.env.truth.get_labels(instance)

            # Set the labels in the active learner
            learner.env.labels.set_labels(instance, *oracle_labels)
            learner.set_as_labeled(instance)

    def call_hooks(self) -> Mapping[str, Sequence[Any]]:
        results = {
            l_key: [hook(self.learners[l_key]) for hook in hooks]
            for l_key, hooks in self.iteration_hooks.items()
        }
        return results

    def iter_increment(self) -> None:
        self.it += 1

    def __call__(self) -> Mapping[str, Mapping[str, bool]]:
        self._retrain()
        self._query_and_label()
        self.main_stats_tracker[self.it] = self._al_stats()
        self.label_stats_tracker[self.it] = self._label_stats()
        self.metrics_tracker[self.it] = self._calculate_metrics()
        stop_result = self.determine_stop()
        self.stop_tracker[self.it] = stop_result
        self.call_hooks()
        self.iter_increment()
        return stop_result



import logging
import pickle
from pathlib import Path
from uuid import uuid4

from instancelib.utils.func import list_unzip

from .analysis.tarplotter import TarExperimentPlotter
from .benchmarking.reviews import benchmark, read_review_dataset
from .configurations import AL_REPOSITORY, FE_REPOSITORY
from .configurations.base import EXPERIMENT_REPOSITORY, STOP_BUILDER_REPOSITORY
from .configurations.catalog import ExperimentCombination
from .utils.func import flatten_dicts
from .utils.io import create_dir_if_not_exists

LOGGER = logging.getLogger(__name__)


def tar_benchmark(
    dataset_path: Path,
    target_path: Path,
    exp_choice: ExperimentCombination,
    pos_label: str,
    neg_label: str,
) -> None:
    LOGGER.info(f"Start Experiment on {dataset_path} with {exp_choice}")
    exp = EXPERIMENT_REPOSITORY[exp_choice]

    # Parse Configuration

    # Retrieve Configuration
    al_config = AL_REPOSITORY[exp.al_configuration]
    fe_config = FE_REPOSITORY[exp.fe_configuration]
    stop_builders = [
        STOP_BUILDER_REPOSITORY[config] for config in exp.stop_builder_configuration
    ]
    initializer = exp.init_configuration
    estimator_dicts, stop_criteria_dicts = list_unzip(
        map(lambda f: f(pos_label, neg_label), stop_builders)
    )
    estimators = flatten_dicts(*estimator_dicts)
    stop_criteria = flatten_dicts(*stop_criteria_dicts)

    # Specify benchmark targets and outputs
    uuid = uuid4()
    target_path = Path(target_path)
    dataset_name = dataset_path.stem
    # File locations for the plotter object
    dataset_dir = target_path / dataset_name
    plot_filename_pkl = dataset_dir / f"run_{uuid}.pkl"
    plot_filename_pdf = dataset_dir / f"run_{uuid}.pdf"

    # Load the dataset
    env = read_review_dataset(dataset_path)
    create_dir_if_not_exists(dataset_dir)
    plot = benchmark(
        env,
        plot_filename_pkl,
        plot_filename_pdf,
        al_config,
        fe_config,
        initializer,
        estimators,
        stop_criteria,
        pos_label,
        neg_label,
        batch_size=exp.batch_size,
        stop_interval=exp.stop_interval,
        estimation_interval=exp.estimation_interval,
    )
    with plot_filename_pkl.open("wb") as fh:
        pickle.dump(plot, fh)
    plot.show(filename=plot_filename_pdf)



from typing import Any, Generic

from ..activelearning import ActiveLearner
from ..analysis.analysis import process_performance
from ..typehints import LT
from .base import AbstractStopCriterion




class RecallStopCriterion(AbstractStopCriterion[LT], Generic[LT]):
    def __init__(self, label: LT, target_recall: float):
        self.label = label
        self.target_recall = target_recall
        self.recall = 0.0

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        self.recall = process_performance(learner, self.label).recall

    @property
    def stop_criterion(self) -> bool:
        if self.recall >= 0 and self.recall <= 1:
            return self.recall >= self.target_recall
        return False









from __future__ import annotations

import collections
import logging
from typing import Any, Callable, Generic, Optional, Sequence, Set, TypeVar
from typing_extensions import Self

from instancelib import Instance

from ..analysis.base import AnnotationStatistics
from ..environment import AbstractEnvironment
from ..typehints import DT, IT, KT, LT, RT, VT
from .base import ActiveLearner

FT = TypeVar("FT")
F = TypeVar("F", bound=Callable[..., Any])

LOGGER = logging.getLogger(__name__)


class PoolBasedAL(
    ActiveLearner[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    """PoolBasedAL specifies the basis for a poolbased Active Learning algorithm.
    Any algorithm that determines an ordering for *n* instances, can be based on
    this class.

    Examples
    --------
    Creation and initialization by attaching an Environment `env`.

    >>> al = PoolBasedAL(env)

    Instances can be sampled as follows

    >>> instance = next(al)

    Or in batch mode by using :func:`itertools.islice`:

    >>> instances = itertools.islice(al, 10) # Change the number to get more instances

    Mark a document as labeled

    >>> al.set_as_labeled(instance)

    Mark a document as unlabeled

    >>> al.set_as_unlabeled(instance)

    Update the ordering

    >>> al.update_ordering()

    Check how many documents are labeled

    >>> al.len_labeled
    """

    sampled: Set[KT]
    """The documents that were sampled during by this learner"""

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        self.env = env
        self.ordering = None
        self._stats = AnnotationStatistics()
        self.sampled: Set[KT] = set()
        self.identifier = identifier

    def _set_ordering(self, ordering: Sequence[KT]) -> None:
        """Set the ordering of the learner and clear the sampled set

        Parameters
        ----------
        ordering : Sequence[KT]
            The new ordering
        """
        LOGGER.info("Setting a new ordering for %s", self.name)
        self.ordering = collections.deque(ordering)  # type: ignore
        self.sampled.clear()

    def update_ordering(self) -> bool:
        ordering = self.env.unlabeled.key_list
        self._set_ordering(ordering)
        return True

    @property
    def has_ordering(self) -> bool:
        """Check if the learner has an ordering

        Returns
        -------
        bool
            True if it has an ordering
        """
        return self.ordering is not None

    @ActiveLearner.iterator_log
    def __next__(self) -> IT:
        """Query the next instance according to the ordering.
        If the instances has already been sampled during the
        current iteration. The sampled set is cleared after
        each :meth:`.update_ordering()` call

        Returns
        -------
        Instance[KT, DT, VT, RT]
            The next instance that should be labeled

        Raises
        ------
        StopIteration
            If there are no instances left that have not been sampled
        NoOrderingException
            If the ordering was not determined


        Examples
        --------
        :meth:`~allib.activelearning.PoolBasedAL.__next__()` is a special method and
        makes this object combined with the method :meth:`~allib.base.ActiveLearner.__iter__()`
        an :term:`iterator`.
        As any iterator, this object can be used as follows:

        >>> # Initialize the object
        >>> al = PoolBasedAL(env)
        >>> # Request the most informative instance
        >>> ins = next(al)
        >>> # Request the 10 most informative instances
        >>> inss = itertools.islice(al, 10)
        """
        if self.ordering is not None:
            try:
                # Pop the most left item from the deque
                key = self.ordering.popleft()
                # If it has already been sampled, repeat
                while key in self.sampled:
                    key = self.ordering.popleft()
            except IndexError:
                # There are no more items left, raise StopIteration so iterator knows it should stop
                raise StopIteration()
            else:
                # Mark the instance as sampled
                self.sampled.add(key)
                # Retrieve the instance from the dataset
                result = self.env.dataset[key]
                return result
        raise StopIteration()

    @ActiveLearner.label_log
    def set_as_labeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        """Mark the instance as labeled

        Parameters
        ----------
        instance : Instance[KT, DT, VT, RT]
            The instance that should be marked as labeled
        """
        self.env.unlabeled.discard(instance)
        self.env.labeled.add(instance)

    def set_as_unlabeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        """Mark the instance as unlabeled

        Parameters
        ----------
        instance : Instance[KT, DT, VT, RT]
            The instance that should be marked as unlabeled
        """
        self.env.labeled.discard(instance)
        self.env.unlabeled.add(instance)

    @classmethod
    def builder(cls, *args: Any, **kwargs: Any) -> Callable[..., Self]:
        def builder_func(
            env: AbstractEnvironment[IT, KT, DT, VT, RT, LT], *_: Any, **__: Any
        ) -> Self:
            return cls(env)

        return builder_func



from ..activelearning import ActiveLearningFactory
from ..environment import EnvironmentFactory
from ..factory import ObjectFactory
from ..feature_extraction import FeatureExtractionFactory
from .catalog import ModuleCatalog as Cat
from .component import Component


class MainFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.attach(FeatureExtractionFactory())
        self.attach(ActiveLearningFactory())
        self.attach(EnvironmentFactory())



import collections
import itertools
from abc import ABC, abstractmethod
from typing import Any, Deque, Generic, TypeVar

import numpy as np  # type: ignore

from ..activelearning import ActiveLearner
from ..activelearning.ensembles import AbstractEnsemble
from ..activelearning.estimator import Estimator
from ..estimation.base import AbstractEstimator
from ..estimation.rcapture import AbundanceEstimator
from ..utils.func import all_equal, intersection
from ..analysis.analysis import process_performance

KT = TypeVar("KT")
VT = TypeVar("VT")
DT = TypeVar("DT")
RT = TypeVar("RT")
LT = TypeVar("LT")

class AbstractStopCriterion(ABC, Generic[LT]):
    @abstractmethod
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        pass

    @property
    @abstractmethod
    def stop_criterion(self) -> bool:
        pass




from distutils.command.build import build
from os import PathLike
from typing import Any, Optional, Union

import numpy as np
import numpy.typing as npt
from instancelib import BaseVectorizer, Environment
from instancelib.labels.encoder import LabelEncoder, SklearnLabelEncoder
from instancelib.machinelearning.autovectorizer import AutoVectorizerClassifier
from instancelib.machinelearning.skdata import SkLearnDataClassifier
from instancelib.machinelearning.sklearn import SkLearnClassifier

from sklearn.base import ClassifierMixin, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import \
    LabelEncoder as SKLabelEncoder  # type: ignore

from ..balancing import BaseBalancer
from ..typehints import DT, IT, KT, LT


class BalancedSklearnVectorClassifier(AutoVectorizerClassifier[IT, KT, LT]):
    def __init__(self, 
                 estimator: Union[ClassifierMixin, Pipeline], 
                 vectorizer: BaseVectorizer[IT], 
                 encoder: LabelEncoder[LT, npt.NDArray[Any], npt.NDArray[Any], npt.NDArray[Any]],
                 balancer: BaseBalancer, 
                 storage_location: "Optional[PathLike[str]]" = None, 
                 filename: "Optional[PathLike[str]]" = None) -> None:
        super().__init__(estimator, vectorizer, encoder, storage_location, filename)
        self.balancer = balancer

    def _fit(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]):
        x_resampled, y_resampled = self.balancer.resample(x_data, y_data)
        return super()._fit(x_resampled, y_resampled)

    @classmethod
    def build(cls, estimator: Union[ClassifierMixin, Pipeline],
                   vectorizer: BaseVectorizer[IT],
                   balancer: BaseBalancer,  
                   env: Environment[IT, KT, Any, npt.NDArray[Any], Any, LT], 
                   storage_location: "Optional[PathLike[str]]" = None, 
                   filename: "Optional[PathLike[str]]" = None
                   ) -> SkLearnClassifier[IT, KT, Any, npt.NDArray[Any], LT]:
        sklearn_encoder: TransformerMixin = SKLabelEncoder()
        il_encoder = SklearnLabelEncoder(sklearn_encoder, env.labels.labelset)    
        return cls(estimator, vectorizer, il_encoder, balancer, storage_location, filename)


class BalancedDataClassifier(SkLearnDataClassifier[IT, KT, DT, LT]):
    def __init__(self, 
                 estimator: Union[ClassifierMixin, Pipeline], 
                 encoder: LabelEncoder[LT, npt.NDArray[Any], npt.NDArray[Any], npt.NDArray[Any]],
                 balancer: BaseBalancer,
                 storage_location: "Optional[PathLike[str]]" = None, 
                 filename: "Optional[PathLike[str]]" = None) -> None:
        super().__init__(estimator, encoder, storage_location, filename)
        self.balancer = balancer

    def _fit(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]):
        x_resampled, y_resampled = self.balancer.resample(x_data, y_data)
        return super()._fit(x_resampled, y_resampled)

    @classmethod
    def build(cls, estimator: Union[ClassifierMixin, Pipeline],
                   balancer: BaseBalancer,  
                   env: Environment[IT, KT, Any, npt.NDArray[Any], Any, LT], 
                   storage_location: "Optional[PathLike[str]]" = None, 
                   filename: "Optional[PathLike[str]]" = None
                   ) -> SkLearnClassifier[IT, KT, Any, npt.NDArray[Any], LT]:
        sklearn_encoder: TransformerMixin = SKLabelEncoder()
        il_encoder = SklearnLabelEncoder(sklearn_encoder, env.labels.labelset)    
        return cls(estimator, il_encoder, balancer, storage_location, filename)



class NoVectorsException(Exception):
    """This exception is thrown when there are no vectors available
    """    
    pass

class NoLabeledDataException(Exception):
    """This exception is thrown when there are no labeled instances available
    """    
    pass

class NotInitializedException(Exception):
    """This exception is returned if the Active Learner has not been 
    initialized. That is, there is no attached Environment, so it 
    cannot sample instances.
    """    
    pass

class NoOrderingException(Exception):
    """This exception is returned if the instances in the `ActiveLearner`
    have not yet been ordered, or establishing an ordering is not possible
    while sampling instances. In this case, no instances can be returned
    and instead this `Exception` is raised.
    """    
    pass



import collections
from math import ceil
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Optional,
    Sequence,
    Tuple,
)

import instancelib as il
import numpy as np
import numpy.typing as npt
from instancelib.typehints import DT, KT, LT, RT, VT

from ..environment.base import AbstractEnvironment
from ..typehints import IT
from ..utils.func import list_unzip
from .autotar import AutoTarLearner


def calc_ap_prior_distribution(
    ranking: Sequence[Tuple[KT, float]]
) -> Sequence[Tuple[KT, float]]:
    keys, _ = list_unzip(ranking)
    N = len(ranking)
    ranks = np.array(range(1, (N + 1)))
    Z = np.sum(np.log(N / ranks))
    pis: npt.NDArray[Any] = 1.0 / Z * np.log(N / ranks)
    return list(zip(keys, pis.tolist()))


class AutoStopLearner(
    AutoTarLearner[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    distributions: Dict[int, Dict[KT, float]]

    dist_fixed: Dict[int, npt.NDArray[Any]]
    cumulative_fo: Dict[int, npt.NDArray[Any]]
    cumulative_so: Dict[int, npt.NDArray[Any]]
    label_vector: Dict[int, npt.NDArray[Any]]
    cumulative_sampled: Dict[int, FrozenSet[KT]]
    key_seq: Sequence[KT]

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        classifier: Callable[
            [AbstractEnvironment[IT, KT, DT, VT, RT, LT]],
            il.AbstractClassifier[
                IT, KT, DT, VT, RT, LT, npt.NDArray[Any], npt.NDArray[Any]
            ],
        ],
        pos_label: LT,
        neg_label: LT,
        k_sample: int,
        batch_size: int,
        *_,
        seed: int = 0,
        prune: bool = True,
        identifier: Optional[str] = None,
        **__
    ) -> None:
        super().__init__(
            env,
            classifier,
            pos_label,
            neg_label,
            k_sample,
            batch_size,
            *_,
            seed=seed,
            identifier=identifier,
            **__
        )

        super().__init__(
            env,
            classifier,
            pos_label,
            neg_label,
            k_sample,
            batch_size,
            identifier=identifier,
            seed=seed,
        )
        self.distributions = dict()
        self.dist_fixed = dict()
        self.cumulative_fo = dict()
        self.cumulative_so = dict()
        self.cumulative_sampled = dict()
        self.key_seq = tuple(self.env.dataset)
        self.label_vector = dict()
        self.dtype = np.float128
        self.prune = prune

    def inclusion_probability(self, key: KT, t_max: int) -> float:
        dist_h = self.distributions
        pi = 1.0 - np.product(
            [
                ((1.0 - dist_h[t][key]) ** len(self.sampled_sets[t]))
                for t in dist_h
                if t <= t_max
            ]
        )
        return pi  # type: ignore

    def second_order_probability(self, key_a: KT, key_b: KT, t_max: int) -> float:
        dist_h = self.distributions
        min_part = 1.0 - np.product(
            [
                (
                    (1.0 - dist_h[t][key_a] - dist_h[t][key_b])
                    ** len(self.sampled_sets[t])
                )
                for t in dist_h
                if t <= t_max
            ]
        )
        pij = (
            self.inclusion_probability(key_a, t_max)
            + self.inclusion_probability(key_b, t_max)
            - min_part
        )
        return pij  # type: ignore

    def _sample(self, distribution: Sequence[Tuple[KT, float]]) -> Sequence[KT]:
        keys, probs = list_unzip(distribution)
        sample: Sequence[KT] = self.rng.choice(
            keys, size=self.batch_size, p=probs  # type: ignore
        ).tolist()
        return sample

    def _update_inclusion_prob(self, it: int):
        # Fixing some variables
        big_n = len(self.env.dataset)
        dists = self.distributions[it]
        prev_it = it - 1

        n = len(self.sampled_sets[it])
        dist = np.array([dists[k] for k in self.key_seq]).reshape((1, -1))
        if prev_it in self.cumulative_fo:
            prev_fo = self.cumulative_fo[prev_it]
            prev_so = self.cumulative_so[prev_it]
        else:
            prev_fo = np.zeros((1, big_n), dtype=self.dtype)
            prev_so = np.zeros((big_n, big_n), dtype=self.dtype)

        # Calculate first order (fo) part
        assert it not in self.cumulative_fo
        self.cumulative_fo[it] = prev_fo + n * np.log(1.0 - dist)

        # Calculate second order (so) part
        M = np.tile(dist, (big_n, 1))
        MT = M.T
        temp = 1.0 - M - MT
        # set diagonal values to 1 to make sure log calculation is valid
        np.fill_diagonal(temp, 1)

        assert it not in self.cumulative_so
        self.cumulative_so[it] = prev_so + n * np.log(temp)
        self.dist_fixed[it] = dist

        # Calculate fast binary label vector
        inclusions = frozenset(
            self.env.get_subset_by_labels(self.env.dataset, self.pos_label)
        )
        self.label_vector[it] = np.array([k in inclusions for k in self.key_seq])

    def fo_inclusion_probabilities(self, it: int) -> npt.NDArray[Any]:
        result = 1.0 - np.exp(self.cumulative_fo[it])
        return result

    def so_inclusion_probabilities(self, it: int) -> npt.NDArray[Any]:
        big_n = len(self.env.dataset)
        fo_prob = self.fo_inclusion_probabilities(it)
        M = np.tile(fo_prob, (big_n, 1))
        MT = M.T
        cu_so = self.cumulative_so[it]
        result = M + MT - (1.0 - np.exp(cu_so))
        return result

    def update_sample(self) -> Deque[KT]:
        if not self.current_sample:
            self.stats.update(self)
            if self.it > 0:  # No previous iteration at the start
                # The previous sample has been finished
                # Update probabilities so we can estimate the number of inclusions
                prev_it = self.it - 1
                self.cumulative_sampled[prev_it] = frozenset(self.env.labeled)
                self._update_inclusion_prob(prev_it)
                if self.prune:
                    pprev_it = prev_it - 1
                    if pprev_it in self.cumulative_fo:
                        del self.cumulative_fo[pprev_it]
                        del self.cumulative_so[pprev_it]

            # Sample a new batch
            self._temp_augment_and_train()
            ranking = self._rank(self.env.dataset)
            distribution = calc_ap_prior_distribution(ranking)
            sample = self._sample(distribution)

            # Store all data for later analysis
            self.distributions[self.it] = dict(distribution)
            self.rank_history[self.it] = self._to_history(ranking)
            self.sampled_sets[self.it] = tuple(sample)
            self.batch_sizes[self.it] = self.batch_size

            # Calculate new sample size for the next iteration
            self.current_sample = collections.deque(sample)
            self.batch_size += ceil(self.batch_size / 10)
            self.it += 1

        return self.current_sample

    def __next__(self) -> IT:
        if len(self.env.unlabeled) != 1:
            return super().__next__()
        return next(iter(self.env.unlabeled.values()))
        


from __future__ import annotations

from dataclasses import dataclass
from typing import Any, FrozenSet, Generic, Mapping, OrderedDict, Sequence, Tuple, Dict

import numpy as np
import numpy.typing as npt

from ..typehints.typevars import LT
from ..utils.func import flatten_dicts

from ..activelearning.base import ActiveLearner
from ..activelearning.ensembles import AbstractEnsemble
from .analysis import process_performance

import instancelib as il

@dataclass
class EstimationModelStatistics:
    beta: npt.NDArray[Any]
    mfit: npt.NDArray[Any]
    deviance: float
    preds: npt.NDArray[Any]

@dataclass
class TarDatasetStats:
    pos_count: int
    neg_count: int
    size: int
    prevalence: float

    @classmethod
    def from_learner(cls, 
                     learner: ActiveLearner[Any, Any, Any, Any, Any, LT],
                     pos_label: LT,
                     neg_label: LT) -> TarDatasetStats:
        pos_count = len(learner.env.get_subset_by_labels(learner.env.dataset, pos_label, labelprovider=learner.env.truth))
        neg_count = len(learner.env.get_subset_by_labels(learner.env.dataset, neg_label, labelprovider=learner.env.truth))
        size = len(learner.env.dataset)
        prevalence = pos_count / size
        return TarDatasetStats(pos_count, neg_count, size, prevalence)

@dataclass
class TemporalRecallStats:
    wss: float
    recall: float
    proportional_effort: float
    pos_docs_found: int
    neg_docs_found: int
    effort: int
    loss_er: float
    child_statistics: Sequence[Tuple[str, TemporalRecallStats]]

    @classmethod
    def from_learner(cls,  
                     learner: ActiveLearner[Any, Any, Any, Any, Any, LT],
                     pos_label: LT,
                     neg_label: LT
                     ) -> TemporalRecallStats:
        perf = process_performance(learner, pos_label)
        pos_docs = len(learner.env.get_subset_by_labels(learner.env.labeled, pos_label))
        neg_docs = len(learner.env.get_subset_by_labels(learner.env.labeled, neg_label))
        effort = len(learner.env.labeled)
        prop_effort = effort / len(learner.env.dataset)
        if isinstance(learner, AbstractEnsemble):
            subresults: Sequence[Tuple[str, TemporalRecallStats]] = tuple([
                (str(learner.name), cls.from_learner(learner, pos_label, neg_label)) for learner  in learner.learners])
        else:
            subresults: Sequence[Tuple[str, TemporalRecallStats]] = tuple([])
        return TemporalRecallStats(perf.wss, perf.recall, prop_effort, pos_docs, neg_docs, effort, perf.loss_er, subresults)
    
    def flatten(self, root_name="root") -> Dict[str, TemporalRecallStats]:
        root = {root_name: self}
        children = flatten_dicts(*[stat.flatten(name) for name, stat in self.child_statistics])
        return {**root, **children}

    @staticmethod
    def transpose_dict(recall_stats: Mapping[int, TemporalRecallStats]
                 ) -> Mapping[str, Mapping[int, TemporalRecallStats]]:
        flattened = {key: stat.flatten() for key, stat in recall_stats.items()}
        if flattened:
            learner_names = list(next(iter(flattened.values())).keys())
            ret = {learner: {t: substat[learner] for t,substat in flattened.items()} for learner in learner_names}
            return ret
        return dict()

@dataclass
class DatasetStats(Generic[LT]):
    labels: FrozenSet[LT]

@dataclass
class ALStats:
    unlabeled: int
    labeled: int
    dataset: int

@dataclass
class LabelALStatistics(Generic[LT]):
    label: LT
    seen: int
    generated: int

@dataclass
class BinaryClassificationStatistics(Generic[LT]):
    label: LT
    recall: float
    precision: float
    accuracy: float
    f1: float



import random
from typing import Generic

from ..typehints import DT, IT, KT, LT, RT, VT
from .poolbased import PoolBasedAL


class RandomSampling(
    PoolBasedAL[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    _name = "Random"

    def update_ordering(self) -> bool:
        keys = list(self.env.unlabeled.key_list)
        random.shuffle(keys)
        self._set_ordering(keys)
        return True



# from __future__ import annotations

# import functools
# from abc import ABC, abstractstaticmethod
# from typing import List, Optional

# import numpy as np
# import numpy.typing as npt
# from sklearn.exceptions import NotFittedError
# from scipy.stats import entropy

# from ..environment import AbstractEnvironment
# from ..instances import Instance
# from ..machinelearning import AbstractClassifier

# from .base import ActiveLearner
# from .ml_based import MLBased
# from .poolbased import PoolbasedAL
# from .random import RandomSampling

# class QBC(PoolbasedAL):
#     _name = "QBC"
#     def __init__(self, classifier_list: List[AbstractClassifier]):
#         super().__init__(None)
#         self.classifiers = classifier_list
#         self.fallback = RandomSampling(classifier_list[0])
#         self.commitee_size = len(classifier_list)

#     def __call__(self, environment: AbstractEnvironment) -> QBC:
#         super().__call__(environment)
#         self.fallback = self.fallback(environment)
#         for classifier in self.classifiers:
#             classifier(environment)
#         return self

#     @staticmethod
#     def selection_criterion(prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
#         return - entropy(prob_vec, axis=0)

#     @MLBased.query_fallback
#     @ActiveLearner.query_log
#     def query(self) -> Optional[Instance]:
#         """Select the instance whose posterior probability is most near 0.5

#         Returns
#         -------
#         Optional[Instance]
#             An object containing:
#                 - The document identifier
#                 - The document vector
#                 - The document data
#         """
#         unlabeled_vectors = self._unlabeled.feature_matrix.matrix
#         commitee_votes = []
#         for classifier in self.classifiers:
#             member_votes = classifier.predict(unlabeled_vectors)
#             commitee_votes.append(member_votes)
#         total_votes = np.array(commitee_votes).sum(axis=0) / self.commitee_size
#         metric_result = self.selection_criterion(total_votes)
#         np_idx = np.argmax(metric_result)
#         doc_id = self._unlabeled.feature_matrix.get_instance_id(np_idx)
#         if doc_id is not None:
#             return self._unlabeled[doc_id]
#         return None

#     @MLBased.query_batch_fallback
#     @ActiveLearner.query_batch_log
#     def query_batch(self, batch_size: int) -> List[Instance]:
#         unlabeled_vectors = self._unlabeled.feature_matrix.matrix
#         commitee_votes = []
#         for classifier in self.classifiers:
#             member_votes = classifier.predict(unlabeled_vectors)
#             commitee_votes.append(member_votes)
#         total_votes = np.array(commitee_votes).sum(axis=0) / self.commitee_size
#         metric_result = self.selection_criterion(total_votes)
#         np_idxs = np.flip(np.argsort(metric_result[np.argpartition(metric_result, -batch_size)])).tolist()
#         results = []
#         for np_idx in np_idxs:
#             doc_id = self._unlabeled.feature_matrix.get_instance_id(np_idx)
#             if doc_id is not None:
#                 results.append(self._unlabeled[doc_id])
#         return results

